import argparse
import functools
import gc
import logging
import os
import sys
import typing

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import polars as pl

from ._add_bool_arg import add_bool_arg
from ._alifestd_mark_leaves_polars import alifestd_mark_leaves_polars
from ._alifestd_prune_extinct_lineages_polars import (
    alifestd_prune_extinct_lineages_polars,
)
from ._alifestd_topological_sensitivity_warned_polars import (
    alifestd_topological_sensitivity_warned_polars,
)
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration
from ._log_memory_usage import log_memory_usage


@alifestd_topological_sensitivity_warned_polars(
    insert=False,
    delete=True,
    update=False,
)
def alifestd_downsample_tips_canopy_polars(
    phylogeny_df: pl.DataFrame,
    num_tips: typing.Optional[int] = None,
    criterion: str = "origin_time",
) -> pl.DataFrame:
    """Retain the `num_tips` leaves with the largest `criterion` values and
    prune extinct lineages.

    If `num_tips` is ``None``, it defaults to the number of leaves that
    share the maximum value of the `criterion` column. If `num_tips` is
    greater than or equal to the number of leaves in the phylogeny, the
    whole phylogeny is returned. Ties are broken arbitrarily.

    Only supports asexual phylogenies.

    Parameters
    ----------
    phylogeny_df : polars.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    num_tips : int, optional
        Number of tips to retain. If ``None``, defaults to the count of
        leaves with the maximum `criterion` value.
    criterion : str, default "origin_time"
        Column name used to rank leaves. The `num_tips` leaves with the
        largest values in this column are retained. Ties are broken
        arbitrarily.

    Raises
    ------
    NotImplementedError
        If `phylogeny_df` has no "ancestor_id" column.
    ValueError
        If `criterion` is not a column in `phylogeny_df`.

    Returns
    -------
    polars.DataFrame
        The pruned phylogeny in alife standard format.

    See Also
    --------
    alifestd_downsample_tips_canopy_asexual :
        Pandas-based implementation.
    """
    logging.info(
        "- alifestd_downsample_tips_canopy_polars: collecting schema...",
    )
    schema_names = phylogeny_df.lazy().collect_schema().names()
    gc.collect()
    log_memory_usage(logging.info)
    if criterion not in schema_names:
        raise ValueError(
            f"criterion column {criterion!r} not found in phylogeny_df",
        )

    if "ancestor_id" not in schema_names:
        raise NotImplementedError("ancestor_id column required")

    logging.info(
        "- alifestd_downsample_tips_canopy_polars: checking empty...",
    )
    if phylogeny_df.lazy().limit(1).collect().is_empty():
        return phylogeny_df

    logging.info(
        "- alifestd_downsample_tips_canopy_polars: finding leaf ids...",
    )
    phylogeny_df = alifestd_mark_leaves_polars(phylogeny_df)
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_canopy_polars: selecting top leaf_ids...",
    )
    leaves_lazy = phylogeny_df.lazy().filter(pl.col("is_leaf"))
    if num_tips is None:
        max_val = leaves_lazy.select(pl.col(criterion).max()).collect().item()
        num_tips = (
            leaves_lazy.filter(pl.col(criterion) == max_val)
            .select(pl.len())
            .collect()
            .item()
        )
        gc.collect()
        log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_canopy_polars: counting leaves...",
    )
    total_leaves = leaves_lazy.select(pl.len()).collect().item()
    logging.info(
        f"- alifestd_downsample_tips_canopy_polars: {total_leaves=}...",
    )

    if num_tips >= total_leaves:
        logging.info(
            "- alifestd_downsample_tips_canopy_polars: taking all...",
        )
        leaf_ids = leaves_lazy.select(pl.col("id")).collect().to_series()
    else:  # split case to prevent extreme top_k crash where num_tips is high
        logging.info(
            "- alifestd_downsample_tips_canopy_polars: taking top k...",
        )
        leaf_ids = (
            leaves_lazy.top_k(num_tips, by=pl.col(criterion))
            .select(pl.col("id"))
            .collect()
            .to_series()
        )
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_canopy_polars: marking extant...",
    )
    phylogeny_df = phylogeny_df.with_columns(
        extant=pl.col("id").is_in(leaf_ids),
    )
    del leaf_ids
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_downsample_tips_canopy_polars: pruning...",
    )
    return alifestd_prune_extinct_lineages_polars(phylogeny_df).drop("extant")


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Retain the `-n` leaves with the largest `--criterion` values and prune extinct lineages.

If `-n` is greater than or equal to the number of leaves in the phylogeny, the whole phylogeny is returned. Ties are broken arbitrarily.

Data is assumed to be in alife standard format.
Only supports asexual phylogenies.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.
Otherwise, no action is taken.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_downsample_tips_canopy_asexual :
    CLI entrypoint for Pandas-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_downsample_tips_canopy_polars",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "-n",
        default=sys.maxsize,
        type=int,
        help="Number of tips to retain.",
    )
    parser.add_argument(
        "--criterion",
        default="origin_time",
        type=str,
        help="Column name used to rank leaves; ties broken arbitrarily (default: origin_time).",
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_downsample_tips_canopy_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=functools.partial(
                    alifestd_downsample_tips_canopy_polars,
                    num_tips=args.n,
                    criterion=args.criterion,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



import argparse
import logging
import os
import typing

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import pandas as pd

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._jit import jit
from ._jit_numba_dict_t import jit_numba_dict_t
from ._jit_numpy_int64_t import jit_numpy_int64_t
from ._log_context_duration import log_context_duration


@jit(nopython=True)
def _reassign_ids_asexual(
    ids: np.ndarray,
    ancestor_ids: np.ndarray,
) -> np.ndarray:
    max_id = np.max(ids)

    if max_id > len(ids) * 5:
        reassignment = jit_numba_dict_t.empty(
            key_type=jit_numpy_int64_t,
            value_type=jit_numpy_int64_t,
        )
        for id_ in ids:
            reassignment[id_] = len(reassignment)
        return np.array(
            [reassignment[ancestor_id] for ancestor_id in ancestor_ids]
        )
    else:
        reassignment = np.empty(max_id + 1, dtype=np.int64)
        for i, id_ in enumerate(ids):
            reassignment[id_] = i
        return np.array(
            [reassignment[ancestor_id] for ancestor_id in ancestor_ids]
        )


@jit(nopython=True)
def _reassign_ids_sexual(ids: np.ndarray) -> typing.Dict[int, int]:
    reassignment = jit_numba_dict_t.empty(
        key_type=jit_numpy_int64_t,
        value_type=jit_numpy_int64_t,
    )
    for id_ in ids:
        reassignment[id_] = len(reassignment)

    return reassignment


def alifestd_assign_contiguous_ids(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Reassign so each organism's id corresponds to its row number.

    Organisms retain the same row location; only id numbers change. Input
    dataframe is not mutated by this operation unless `mutate` True.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    original_ids = phylogeny_df["id"].to_numpy(dtype=np.int64).copy()
    phylogeny_df["id"] = np.arange(len(phylogeny_df))

    if "ancestor_id" in phylogeny_df.columns:
        phylogeny_df["ancestor_id"] = _reassign_ids_asexual(
            original_ids, phylogeny_df["ancestor_id"].to_numpy(dtype=np.uint64)
        )
        if "ancestor_list" in phylogeny_df:
            phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
                phylogeny_df["id"], phylogeny_df["ancestor_id"]
            )
    else:
        reassignment = _reassign_ids_sexual(original_ids)
        phylogeny_df["ancestor_list"] = phylogeny_df["ancestor_list"].map(
            lambda ancestor_list_str: str(
                [
                    int(reassignment[ancestor_id])
                    for ancestor_id in alifestd_parse_ancestor_ids(
                        ancestor_list_str
                    )
                ]
            )
        )
        phylogeny_df.loc[
            phylogeny_df["ancestor_list"] == "[]", "ancestor_list"
        ] = "[none]"

    return phylogeny_df


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Reassign so each organism's id corresponds to its row number.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_assign_contiguous_ids",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_assign_contiguous_ids", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_assign_contiguous_ids,
            ),
        )



from ._assemblage_to_ascii import assemblage_to_ascii
from ._col_to_ascii import col_to_ascii

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "assemblage_to_ascii",
    "col_to_ascii",
]



import argparse
import functools
import logging
import os
import typing

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import polars as pl

from ._add_bool_arg import add_bool_arg
from ._alifestd_coarsen_dilate_asexual import _alifestd_coarsen_dilate_impl
from ._alifestd_has_contiguous_ids_polars import (
    alifestd_has_contiguous_ids_polars,
)
from ._alifestd_is_topologically_sorted_polars import (
    alifestd_is_topologically_sorted_polars,
)
from ._alifestd_mark_leaves_polars import alifestd_mark_leaves_polars
from ._alifestd_topological_sensitivity_warned_polars import (
    alifestd_topological_sensitivity_warned_polars,
)
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@alifestd_topological_sensitivity_warned_polars(
    insert=False,
    delete=True,
    update=True,
)
def alifestd_coarsen_dilate_polars(
    phylogeny_df: typing.Union[pl.DataFrame, pl.LazyFrame],
    *,
    criterion: str = "origin_time",
    dilation: int = 1,
) -> pl.DataFrame:
    """Coarsen a phylogeny by collapsing inner nodes within dilation windows.

    All inner (non-leaf) nodes with criterion values in the half-open
    interval ``[n, n + dilation)``, where ``n % dilation == 0``, are
    collapsed to a single inner node at ``n``.

    Tip nodes are never moved. The MRCA of two tips may only shift
    backward (never forward), by at most ``dilation`` units, and never
    across a ``n % dilation == 0`` boundary.

    Parameters
    ----------
    phylogeny_df : polars.DataFrame or polars.LazyFrame
        Input phylogeny in alife standard format.
    criterion : str, default "origin_time"
        Column whose values define the time axis for dilation.
    dilation : int
        Width of the dilation window.  Must be a positive integer.

    Returns
    -------
    polars.DataFrame
        Coarsened phylogeny in alife standard format.

    Raises
    ------
    NotImplementedError
        If input is not topologically sorted with contiguous ids.
    ValueError
        If *dilation* is not a positive integer, if *criterion* is not
        present in *phylogeny_df*, or if *criterion* is ``"id"`` or
        ``"ancestor_id"``.

    See Also
    --------
    alifestd_coarsen_dilate_asexual :
        Pandas-based implementation.
    """
    if dilation <= 0:
        raise ValueError(f"dilation must be positive, got {dilation}")

    if criterion in ("id", "ancestor_id"):
        raise ValueError(
            f"criterion must not be 'id' or 'ancestor_id', got {criterion!r}",
        )

    logging.info(
        "- alifestd_coarsen_dilate_polars: collecting schema...",
    )
    schema_names = phylogeny_df.lazy().collect_schema().names()

    if criterion not in schema_names:
        raise ValueError(
            f"criterion column {criterion!r} not found in phylogeny_df",
        )

    if "ancestor_list" in schema_names:
        raise NotImplementedError(
            "ancestor_list column not supported in polars implementation",
        )

    if "ancestor_id" not in schema_names:
        raise ValueError("asexual phylogeny required")

    logging.info(
        "- alifestd_coarsen_dilate_polars: checking empty...",
    )
    if phylogeny_df.lazy().limit(1).collect().is_empty():
        return phylogeny_df.lazy().collect()

    logging.info(
        "- alifestd_coarsen_dilate_polars: collecting original ids...",
    )
    original_ids = (
        phylogeny_df.lazy().select("id").collect().to_series().to_numpy()
    )

    logging.info(
        "- alifestd_coarsen_dilate_polars: checking contiguous ids...",
    )
    if not alifestd_has_contiguous_ids_polars(phylogeny_df):
        raise NotImplementedError(
            "alifestd_coarsen_dilate_polars requires contiguous ids",
        )

    logging.info(
        "- alifestd_coarsen_dilate_polars: checking topological sort...",
    )
    if not alifestd_is_topologically_sorted_polars(phylogeny_df):
        raise NotImplementedError(
            "alifestd_coarsen_dilate_polars requires topologically "
            "sorted input",
        )

    logging.info(
        "- alifestd_coarsen_dilate_polars: collecting ancestor_ids...",
    )
    ancestor_ids = (
        phylogeny_df.lazy()
        .select("ancestor_id")
        .collect()
        .to_series()
        .to_numpy()
        .copy()
    )

    logging.info(
        "- alifestd_coarsen_dilate_polars: collecting criterion values...",
    )
    criterion_values = (
        phylogeny_df.lazy()
        .select(criterion)
        .collect()
        .to_series()
        .to_numpy()
        .copy()
    )

    logging.info(
        "- alifestd_coarsen_dilate_polars: marking leaves...",
    )
    phylogeny_df_with_leaves = alifestd_mark_leaves_polars(phylogeny_df)
    is_leaf = (
        phylogeny_df_with_leaves.lazy()
        .select("is_leaf")
        .collect()
        .to_series()
        .to_numpy()
    )

    logging.info(
        "- alifestd_coarsen_dilate_polars: running coarsen dilate...",
    )
    (
        new_ancestor_ids,
        new_criterion_values,
        keep_mask,
    ) = _alifestd_coarsen_dilate_impl(
        ancestor_ids,
        criterion_values,
        is_leaf,
        dilation,
    )

    logging.info(
        "- alifestd_coarsen_dilate_polars: applying results...",
    )
    return (
        phylogeny_df.lazy()
        .collect()
        .filter(keep_mask)
        .with_columns(
            id=original_ids[keep_mask],
            ancestor_id=original_ids[new_ancestor_ids[keep_mask]],
            **{criterion: new_criterion_values[keep_mask]},
        )
    )


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Coarsen a phylogeny by collapsing inner nodes within dilation windows.

All inner (non-leaf) nodes with criterion values in [n, n + dilation),
where n mod dilation == 0, are collapsed to a single inner node at n.

Tip nodes are never moved.

Data is assumed to be in alife standard format.
Only supports asexual phylogenies.

Additional Notes
================
- Requires topologically sorted input with contiguous ids.

- Requires 'ancestor_id' column to be present in input DataFrame.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_coarsen_dilate_asexual :
    CLI entrypoint for Pandas-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_coarsen_dilate_polars",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--criterion",
        default="origin_time",
        type=str,
        help="Column whose values define the time axis (default: origin_time).",
    )
    parser.add_argument(
        "--dilation",
        default=1,
        type=int,
        help="Width of the dilation window (default: 1). Note: with "
        "floating-point criterion values, dilation=1 floors inner-node "
        "values to integers.",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_coarsen_dilate_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=functools.partial(
                    alifestd_coarsen_dilate_polars,
                    criterion=args.criterion,
                    dilation=args.dilation,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                ),
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



import argparse
import logging
import os
import pathlib

import numpy as np
import polars as pl

from ._alifestd_from_newick import _extract_labels, _parse_newick
from ._configure_prod_logging import configure_prod_logging
from ._eval_kwargs import eval_kwargs
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


# Performance: see _alifestd_from_newick.py for benchmark numbers.
def alifestd_from_newick_polars(
    newick: str,
    *,
    branch_length_dtype: type = float,
    create_ancestor_list: bool = False,
) -> pl.DataFrame:
    """Convert a Newick format string to a phylogeny dataframe.

    Parses a Newick tree string and returns a polars DataFrame in alife
    standard format with columns: id, ancestor_id, taxon_label,
    origin_time_delta, and branch_length. Optionally includes
    ancestor_list.

    Parameters
    ----------
    newick : str
        A phylogeny in Newick format.
    branch_length_dtype : type, default float
        Dtype for branch length values. Use ``int`` to get nullable integer
        columns (``pl.Int64``). Missing branch lengths will be ``null`` for
        integer dtypes or ``NaN`` for float dtypes.
    create_ancestor_list : bool, default False
        If True, include an ``ancestor_list`` column in the result.

    Returns
    -------
    pl.DataFrame
        Phylogeny dataframe in alife standard format.

    See Also
    --------
    alifestd_from_newick :
        Pandas-based implementation.
    alifestd_as_newick_asexual :
        Inverse conversion, from alife standard to Newick format.
    """
    newick = newick.strip()
    if not newick:
        columns = {
            "id": pl.Series([], dtype=pl.Int64),
            "ancestor_id": pl.Series([], dtype=pl.Int64),
            "taxon_label": pl.Series([], dtype=pl.Utf8),
            "origin_time_delta": pl.Series([], dtype=pl.Float64),
            "branch_length": pl.Series([], dtype=pl.Float64),
        }
        if create_ancestor_list:
            columns["ancestor_list"] = pl.Series([], dtype=pl.Utf8)
        return pl.DataFrame(columns)

    chars = np.frombuffer(newick.encode("ascii"), dtype=np.uint8)
    n = len(chars)

    (
        ids,
        ancestor_ids,
        branch_lengths,
        label_start_stops,
    ) = _parse_newick(newick, chars, n, branch_length_dtype)

    labels = _extract_labels(newick, chars, label_start_stops)

    # convert branch lengths to requested dtype with proper null handling
    np_dtype = np.dtype(branch_length_dtype)
    if np.issubdtype(np_dtype, np.integer):
        # NaN -> null, then cast to Int64
        bl_series = pl.Series(branch_lengths).fill_nan(None).cast(pl.Int64)
        otd_series = bl_series.clone()
    else:
        bl_series = pl.Series(branch_lengths)
        otd_series = pl.Series(branch_lengths.copy())

    columns = {
        "id": pl.Series(ids, dtype=pl.Int64),
        "ancestor_id": pl.Series(ancestor_ids, dtype=pl.Int64),
        "taxon_label": pl.Series(labels.tolist(), dtype=pl.Utf8),
        "origin_time_delta": otd_series,
        "branch_length": bl_series,
    }

    if create_ancestor_list:
        ancestor_list = []
        for id_, anc_id in zip(ids, ancestor_ids):
            if id_ == anc_id:
                ancestor_list.append("[none]")
            else:
                ancestor_list.append(f"[{anc_id}]")
        columns["ancestor_list"] = pl.Series(ancestor_list, dtype=pl.Utf8)

    return pl.DataFrame(columns)


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()})

Convert Newick format phylogeny data to Alife standard format (Polars).

Note that this CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--input-file",
        type=str,
        help="Newick file to convert to Alife standard dataframe format.",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        help="Path to write Alife standard dataframe output to.",
    )
    parser.add_argument(
        "--branch-length-dtype",
        type=str,
        choices=["float", "int"],
        default="float",
        help="Dtype for branch length values. Defaults to 'float'.",
    )
    parser.add_argument(
        "--create-ancestor-list",
        action="store_true",
        default=False,
        help="Include an ancestor_list column in the output.",
    )
    parser.add_argument(
        "--output-kwarg",
        action="append",
        dest="output_kwargs",
        type=str,
        default=[],
        help=(
            "Additional keyword arguments to pass to output engine call. "
            "Provide as 'key=value'. "
            "Specify multiple kwargs by using this flag multiple times. "
            "Arguments will be evaluated as Python expressions. "
            "Example: 'include_header=False'"
        ),
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=get_hstrat_version(),
    )
    return parser


_dtype_lookup = {"float": float, "int": int}


if __name__ == "__main__":
    configure_prod_logging()

    parser = _create_parser()
    args = parser.parse_args()

    logging.info(f"reading Newick data from {args.input_file}...")
    newick_str = pathlib.Path(args.input_file).read_text()

    with log_context_duration(
        "hstrat._auxiliary_lib.alifestd_from_newick_polars", logging.info
    ):
        logging.info("converting from Newick format...")
        phylogeny_df = alifestd_from_newick_polars(
            newick_str,
            branch_length_dtype=_dtype_lookup[args.branch_length_dtype],
            create_ancestor_list=args.create_ancestor_list,
        )

    output_ext = os.path.splitext(args.output_file)[1]
    output_kwargs = eval_kwargs(args.output_kwargs)
    dispatch_writer = {
        ".csv": pl.DataFrame.write_csv,
        ".fea": pl.DataFrame.write_ipc,
        ".feather": pl.DataFrame.write_ipc,
        ".pqt": pl.DataFrame.write_parquet,
        ".parquet": pl.DataFrame.write_parquet,
    }

    logging.info(
        f"writing alife-standard {output_ext} phylogeny data to "
        f"{args.output_file}...",
    )
    dispatch_writer[output_ext](
        phylogeny_df,
        args.output_file,
        **output_kwargs,
    )

    logging.info("done!")



import math
import os

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_leaf_ids,
    alifestd_find_root_ids,
    alifestd_has_multiple_roots,
    alifestd_make_empty,
    alifestd_mark_colless_like_index_sd_asexual,
    alifestd_mark_num_children_asexual,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)
from hstrat._auxiliary_lib._alifestd_mark_colless_like_index_mdm_asexual import (
    _colless_like_fast_path,
    _colless_like_slow_path,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_mark_colless_like_index_sd_asexual(phylogeny_df)

    assert alifestd_validate(result)
    assert original.equals(phylogeny_df)

    assert all(result["colless_like_index_sd"] >= 0)

    leaf_ids = [*alifestd_find_leaf_ids(phylogeny_df)]
    for leaf_id in leaf_ids:
        val = result[result["id"] == leaf_id][
            "colless_like_index_sd"
        ].squeeze()
        assert val == 0.0

    if not alifestd_has_multiple_roots(phylogeny_df):
        (root_id,) = alifestd_find_root_ids(phylogeny_df)
        root_val = result[result["id"] == root_id][
            "colless_like_index_sd"
        ].squeeze()
        assert root_val >= 0
        assert root_val == result["colless_like_index_sd"].max()


def test_empty():
    res = alifestd_mark_colless_like_index_sd_asexual(
        alifestd_make_empty(),
    )
    assert "colless_like_index_sd" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_chain(mutate: bool):
    """Chain: all unifurcating -> balance = 0."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[0, "colless_like_index_sd"] == 0.0
    assert result_df.loc[1, "colless_like_index_sd"] == 0.0
    assert result_df.loc[2, "colless_like_index_sd"] == 0.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_bifurcating_balanced(mutate: bool):
    r"""Balanced bifurcating tree: sd(1, 1) = 0.

          0
         / \
        1   2
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[0, "colless_like_index_sd"] == pytest.approx(
        0.0,
    )
    assert result_df.loc[1, "colless_like_index_sd"] == 0.0
    assert result_df.loc[2, "colless_like_index_sd"] == 0.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_bifurcating_imbalanced(mutate: bool):
    r"""Imbalanced bifurcating tree.

          0
         / \
        1   2
           / \
          3   4

    For k=2 values a, b:
        sd = |a - b| / sqrt(2)
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[None]", "[0]", "[0]", "[2]", "[2]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[1, "colless_like_index_sd"] == 0.0
    assert result_df.loc[3, "colless_like_index_sd"] == 0.0
    assert result_df.loc[4, "colless_like_index_sd"] == 0.0
    assert result_df.loc[2, "colless_like_index_sd"] == pytest.approx(
        0.0,
    )

    # sd(a, b) = |a - b| / sqrt(2)
    fsize_1 = math.log(0 + math.e)
    fsize_2 = math.log(2 + math.e) + 2 * math.log(0 + math.e)
    expected_sd = abs(fsize_2 - fsize_1) / math.sqrt(2)
    assert result_df.loc[0, "colless_like_index_sd"] == pytest.approx(
        expected_sd,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_polytomy_balanced(mutate: bool):
    r"""Balanced polytomy: sd(1, 1, 1) = 0.

          0
        / | \
       1  2  3
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[0]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    assert result_df.loc[0, "colless_like_index_sd"] == pytest.approx(
        0.0,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_polytomy_imbalanced(mutate: bool):
    r"""Imbalanced polytomy.

            0
          / | \
         1  2  3
              / \
             4   5

    Children of 0: f-sizes = [1.0, 1.0, ln(2+e)+2.0]
    sd = sqrt(var) = sqrt((1/(3-1)) * sum (x_i - mean)^2)
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[3]",
                "[3]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    for leaf in [1, 2, 4, 5]:
        assert result_df.loc[leaf, "colless_like_index_sd"] == 0.0

    assert result_df.loc[3, "colless_like_index_sd"] == pytest.approx(
        0.0,
    )

    fsize_3 = math.log(2 + math.e) + 2 * math.log(0 + math.e)
    vals = [1.0, 1.0, fsize_3]
    mean = sum(vals) / 3
    expected_sd = math.sqrt(sum((v - mean) ** 2 for v in vals) / 2)
    assert result_df.loc[0, "colless_like_index_sd"] == pytest.approx(
        expected_sd,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_non_contiguous_ids(mutate: bool):
    """Non-contiguous IDs to exercise slow path."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [10, 20, 30, 40, 50],
            "ancestor_list": [
                "[None]",
                "[10]",
                "[10]",
                "[30]",
                "[30]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    fsize_20 = math.log(0 + math.e)
    fsize_30 = math.log(2 + math.e) + 2 * math.log(0 + math.e)
    expected_sd = abs(fsize_30 - fsize_20) / math.sqrt(2)
    assert result_df.loc[10, "colless_like_index_sd"] == pytest.approx(
        expected_sd,
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_fast_slow_path_agreement(mutate: bool):
    """Fast and slow paths should produce the same values."""
    phylogeny_df_nc = pd.DataFrame(
        {
            "id": [10, 20, 30, 40, 50, 60, 70],
            "ancestor_list": [
                "[None]",
                "[10]",
                "[10]",
                "[10]",
                "[10]",
                "[50]",
                "[50]",
            ],
        }
    )
    phylogeny_df_c = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[0]",
                "[4]",
                "[4]",
            ],
        }
    )
    result_nc = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df_nc,
        mutate=mutate,
    )
    result_c = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df_c,
        mutate=mutate,
    )

    result_nc.index = result_nc["id"]
    result_c.index = result_c["id"]
    assert result_nc.loc[10, "colless_like_index_sd"] == pytest.approx(
        result_c.loc[0, "colless_like_index_sd"],
    )


def test_symmetric_tree_is_zero():
    """Fully symmetric trees should have Colless-like sd index 0."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
        }
    )
    result_df = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
    )
    assert result_df["colless_like_index_sd"].to_list() == pytest.approx(
        [0.0] * len(result_df),
    )


def test_relationship_sd_var():
    """sd should be sqrt(var) at every node."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[3]",
                "[3]",
            ],
        }
    )
    from hstrat._auxiliary_lib import (
        alifestd_mark_colless_like_index_var_asexual,
    )

    result_sd = alifestd_mark_colless_like_index_sd_asexual(
        phylogeny_df,
    )
    result_var = alifestd_mark_colless_like_index_var_asexual(
        phylogeny_df,
    )

    # For each node the accumulated sd should relate to accumulated
    # var at the root. For a single internal node with k>1 children,
    # sd = sqrt(var). But because these are accumulated (summed),
    # the relationship holds per-node balance, not per accumulated
    # value. Verify at root for this simple tree that has only one
    # non-trivial internal node (node 0).
    result_sd.index = result_sd["id"]
    result_var.index = result_var["id"]

    sd_root = result_sd.loc[0, "colless_like_index_sd"]
    var_root = result_var.loc[0, "colless_like_index_var"]
    assert sd_root == pytest.approx(math.sqrt(var_root))


def test_fast_slow_path_direct_comparison():
    """Directly call fast and slow paths and compare results."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[0]",
                "[4]",
                "[4]",
            ],
        }
    )
    phylogeny_df = alifestd_try_add_ancestor_id_col(
        phylogeny_df,
        mutate=True,
    )
    phylogeny_df = alifestd_mark_num_children_asexual(
        phylogeny_df,
        mutate=True,
    )
    fast_result = _colless_like_fast_path(
        phylogeny_df["ancestor_id"].to_numpy(),
        2,
    )
    slow_result = _colless_like_slow_path(phylogeny_df.copy(), "sd")

    np.testing.assert_allclose(fast_result, slow_result)



import argparse
import functools
import logging
import os

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._add_bool_arg import add_bool_arg
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_mark_num_children_asexual import (
    alifestd_mark_num_children_asexual,
)
from ._alifestd_mark_roots import alifestd_mark_roots
from ._alifestd_topological_sensitivity_warned import (
    alifestd_topological_sensitivity_warned,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@alifestd_topological_sensitivity_warned(
    insert=False,
    delete=True,
    update=True,
)
def alifestd_delete_unifurcating_roots_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    root_ancestor_token: str = "none",
) -> pd.DataFrame:
    """Pare record to bypass root nodes with only one descendant.

    Dataframe reindexing (e.g., df.index) may be applied.

    See also `alifestd_collapse_unifurcations`.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    "[]". Default "none".

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly asexual")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if "num_children" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_num_children_asexual(
            phylogeny_df, mutate=True
        )
    if "is_root" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_roots(phylogeny_df, mutate=True)

    phylogeny_df["is_unifurcating_root"] = (
        phylogeny_df["num_children"] == 1
    ) & phylogeny_df["is_root"]

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.set_index("id", drop=False, inplace=True)

    phylogeny_df["ancestor_is_unifurcating_root"] = phylogeny_df.loc[
        phylogeny_df["ancestor_id"].values, "is_unifurcating_root"
    ].values

    phylogeny_df.loc[
        phylogeny_df["ancestor_is_unifurcating_root"].values,
        "ancestor_id",
    ] = phylogeny_df.loc[
        phylogeny_df["ancestor_is_unifurcating_root"],
        "id",
    ].values

    phylogeny_df = phylogeny_df[~phylogeny_df["is_unifurcating_root"]].copy()

    if "ancestor_list" in phylogeny_df:
        phylogeny_df.loc[:, "ancestor_list"] = alifestd_make_ancestor_list_col(
            phylogeny_df["id"],
            phylogeny_df["ancestor_id"],
            root_ancestor_token=root_ancestor_token,
        )

    return phylogeny_df


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Pare record to bypass root nodes with only one descendant.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_delete_unifurcating_roots_asexual",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--root-ancestor-token",
        type=str,
        default="none",
        help='token for root ancestor entries (default: "none")',
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_delete_unifurcating_roots_asexual",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                functools.partial(
                    alifestd_delete_unifurcating_roots_asexual,
                    root_ancestor_token=args.root_ancestor_token,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            ),
        )



import pytest

from hstrat._auxiliary_lib import eval_kwargs


def test_empty():
    assert eval_kwargs([]) == {}


def test_single_string():
    assert eval_kwargs(["key='value'"]) == {"key": "value"}


def test_single_int():
    assert eval_kwargs(["x=42"]) == {"x": 42}


def test_single_float():
    assert eval_kwargs(["origin_time=0.0"]) == {"origin_time": 0.0}


def test_single_none():
    assert eval_kwargs(["infer_schema_length=None"]) == {
        "infer_schema_length": None,
    }


def test_multiple():
    result = eval_kwargs(["x=1", "y='hello'", "z=None"])
    assert result == {"x": 1, "y": "hello", "z": None}


def test_bool():
    assert eval_kwargs(["flag=True"]) == {"flag": True}
    assert eval_kwargs(["flag=False"]) == {"flag": False}


def test_invalid_raises():
    with pytest.raises(ValueError, match="Failed to parse"):
        eval_kwargs(["not valid python!!!"])



import os
import typing
import warnings

import pandas as pd

from ._alifestd_check_topological_sensitivity import (
    alifestd_check_topological_sensitivity,
)


def _alifestd_warn_topological_sensitivity(
    invalidated: typing.List[str],
    caller: str,
    *,
    insert: bool,
    delete: bool,
    update: bool,
) -> None:
    """Private helper: emit a warning listing invalidated columns."""
    if (
        invalidated
        and "HSTRAT_ALIFESTD_WARN_TOPOLOGICAL_SENSITIVITY_SUPPRESS"
        not in os.environ
    ):
        ops = "/".join(
            name
            for flag, name in [
                (insert, "insert"),
                (delete, "delete"),
                (update, "update"),
            ]
            if flag
        )
        warnings.warn(
            f"{caller} performs {ops} operations that do not update "
            f"topology-dependent columns, which may be invalidated: "
            f"{invalidated}. "
            "Use `origin_time` to recalculate branch lengths for "
            "collapsed phylogeny. To silence this warning, use "
            "alifestd_drop_topological_sensitivity{_polars} or set "
            "HSTRAT_ALIFESTD_WARN_TOPOLOGICAL_SENSITIVITY_SUPPRESS "
            "environment variable."
        )


def alifestd_warn_topological_sensitivity(
    phylogeny_df: pd.DataFrame,
    caller: str,
    *,
    insert: bool,
    delete: bool,
    update: bool,
) -> None:
    """Emit a warning if `phylogeny_df` contains columns that may be
    invalidated by topological operations.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.
    caller : str
        Name of the calling function, included in the warning message.
    insert : bool
        Whether the operation inserts new nodes.
    delete : bool
        Whether the operation deletes nodes.
    update : bool
        Whether the operation updates ancestor relationships.

    Input dataframe is not mutated by this operation.

    See Also
    --------
    alifestd_warn_topological_sensitivity_polars :
        Polars-based implementation.
    """
    _alifestd_warn_topological_sensitivity(
        alifestd_check_topological_sensitivity(
            phylogeny_df,
            insert=insert,
            delete=delete,
            update=update,
        ),
        caller,
        insert=insert,
        delete=delete,
        update=update,
    )



import os
import pathlib

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_as_newick_asexual,
    alifestd_from_newick,
    alifestd_try_add_ancestor_id_col,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    result = alifestd_from_newick("")
    assert len(result) == 0
    assert "id" in result.columns
    assert "ancestor_list" not in result.columns
    assert "ancestor_id" in result.columns
    assert "taxon_label" in result.columns
    assert "origin_time_delta" in result.columns
    assert "branch_length" in result.columns

    result_with_al = alifestd_from_newick("", create_ancestor_list=True)
    assert "ancestor_list" in result_with_al.columns


def test_just_root():
    result = alifestd_from_newick("root;", create_ancestor_list=True)
    assert len(result) == 1
    assert result["id"].iloc[0] == 0
    assert result["ancestor_id"].iloc[0] == 0  # root is own ancestor
    assert result["taxon_label"].iloc[0] == "root"
    assert result["ancestor_list"].iloc[0] == "[none]"


def test_just_root_no_label():
    result = alifestd_from_newick(";")
    assert len(result) == 1
    assert result["id"].iloc[0] == 0
    assert result["ancestor_id"].iloc[0] == 0
    assert result["taxon_label"].iloc[0] == ""


def test_onlychild():
    result = alifestd_from_newick("(child)root;")
    assert len(result) == 2

    root = result[result["taxon_label"] == "root"]
    child = result[result["taxon_label"] == "child"]

    assert len(root) == 1
    assert len(child) == 1
    assert root["ancestor_id"].iloc[0] == root["id"].iloc[0]
    assert child["ancestor_id"].iloc[0] == root["id"].iloc[0]


def test_twins():
    result = alifestd_from_newick("(twin1,twin2)root;")
    assert len(result) == 3

    root = result[result["taxon_label"] == "root"]
    twin1 = result[result["taxon_label"] == "twin1"]
    twin2 = result[result["taxon_label"] == "twin2"]

    assert len(root) == 1
    assert len(twin1) == 1
    assert len(twin2) == 1
    assert twin1["ancestor_id"].iloc[0] == root["id"].iloc[0]
    assert twin2["ancestor_id"].iloc[0] == root["id"].iloc[0]


def test_triplets():
    result = alifestd_from_newick("(triplet1,triplet2,triplet3)root;")
    assert len(result) == 4

    root = result[result["taxon_label"] == "root"]
    for name in ["triplet1", "triplet2", "triplet3"]:
        child = result[result["taxon_label"] == name]
        assert len(child) == 1
        assert child["ancestor_id"].iloc[0] == root["id"].iloc[0]


def test_grandchild():
    result = alifestd_from_newick("((child)parent)root;")
    assert len(result) == 3

    root = result[result["taxon_label"] == "root"]
    parent = result[result["taxon_label"] == "parent"]
    child = result[result["taxon_label"] == "child"]

    assert root["ancestor_id"].iloc[0] == root["id"].iloc[0]
    assert parent["ancestor_id"].iloc[0] == root["id"].iloc[0]
    assert child["ancestor_id"].iloc[0] == parent["id"].iloc[0]


def test_grandchild_and_aunt():
    result = alifestd_from_newick("((child)parent,aunt)root;")
    assert len(result) == 4

    root = result[result["taxon_label"] == "root"]
    parent = result[result["taxon_label"] == "parent"]
    child = result[result["taxon_label"] == "child"]
    aunt = result[result["taxon_label"] == "aunt"]

    assert parent["ancestor_id"].iloc[0] == root["id"].iloc[0]
    assert child["ancestor_id"].iloc[0] == parent["id"].iloc[0]
    assert aunt["ancestor_id"].iloc[0] == root["id"].iloc[0]


def test_branch_lengths():
    result = alifestd_from_newick("(ant:17,(bat:31,cow:22):7,dog:22);")
    # root + ant + internal + bat + cow + dog = 6 nodes
    assert len(result) == 6

    ant = result[result["taxon_label"] == "ant"]
    bat = result[result["taxon_label"] == "bat"]
    cow = result[result["taxon_label"] == "cow"]
    dog = result[result["taxon_label"] == "dog"]

    assert ant["branch_length"].iloc[0] == pytest.approx(17.0)
    assert bat["branch_length"].iloc[0] == pytest.approx(31.0)
    assert cow["branch_length"].iloc[0] == pytest.approx(22.0)
    assert dog["branch_length"].iloc[0] == pytest.approx(22.0)

    # internal node with branch length 7
    internal = result[
        (result["taxon_label"] == "") & (result["ancestor_id"] != result["id"])
    ]
    assert len(internal) == 1
    assert internal["branch_length"].iloc[0] == pytest.approx(7.0)

    # origin_time_delta matches branch_length
    assert (
        result["origin_time_delta"].dropna()
        == result["branch_length"].dropna()
    ).all()


def test_branch_lengths_float():
    result = alifestd_from_newick("(A:0.1,B:0.2,(C:0.3,D:0.4):0.5);")
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    c = result[result["taxon_label"] == "C"]
    d = result[result["taxon_label"] == "D"]

    assert a["branch_length"].iloc[0] == pytest.approx(0.1)
    assert b["branch_length"].iloc[0] == pytest.approx(0.2)
    assert c["branch_length"].iloc[0] == pytest.approx(0.3)
    assert d["branch_length"].iloc[0] == pytest.approx(0.4)


def test_no_branch_length_is_nan():
    result = alifestd_from_newick("(A,B)C;")
    assert np.isnan(result["branch_length"].iloc[0])  # root C
    assert np.isnan(result["branch_length"].iloc[1])  # A
    assert np.isnan(result["branch_length"].iloc[2])  # B


def test_mixed_branch_lengths():
    result = alifestd_from_newick("(A:5,B)C;")
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    assert a["branch_length"].iloc[0] == pytest.approx(5.0)
    assert np.isnan(b["branch_length"].iloc[0])


def test_ancestor_list_col():
    result = alifestd_from_newick("(A,B)C;", create_ancestor_list=True)
    root = result[result["taxon_label"] == "C"]
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]

    assert root["ancestor_list"].iloc[0] == "[none]"
    assert a["ancestor_list"].iloc[0] == f"[{root['id'].iloc[0]}]"
    assert b["ancestor_list"].iloc[0] == f"[{root['id'].iloc[0]}]"

    result_no_al = alifestd_from_newick("(A,B)C;")
    assert "ancestor_list" not in result_no_al.columns


def test_example_newick():
    newick = "(ant:17, (bat:31, cow:22):7, dog:22, (elk:33, fox:12):40);"
    result = alifestd_from_newick(newick)
    # root + ant + internal1 + bat + cow + dog + internal2 + elk + fox = 9
    assert len(result) == 9

    fox = result[result["taxon_label"] == "fox"]
    assert fox["branch_length"].iloc[0] == pytest.approx(12.0)

    elk = result[result["taxon_label"] == "elk"]
    assert elk["branch_length"].iloc[0] == pytest.approx(33.0)


def test_quoted_labels():
    result = alifestd_from_newick("('node A','node B')'root node';")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "node A" in labels
    assert "node B" in labels
    assert "root node" in labels


def test_contiguous_ids():
    result = alifestd_from_newick(
        "(((grandchild1, grandchild2)triplet1,triplet2,triplet3)parent,aunt,uncle)root;"
    )
    assert list(result["id"]) == list(range(len(result)))


@pytest.mark.parametrize(
    "newick_file",
    [
        "grandchild.newick",
        "grandchild_and_aunt.newick",
        "grandchild_and_auntuncle.newick",
        "grandtriplets.newick",
        "grandtriplets_and_aunt.newick",
        "grandtriplets_and_auntuncle.newick",
        "grandtwins.newick",
        "grandtwins_and_aunt.newick",
        "grandtwins_and_auntuncle.newick",
        "greatgrandtwins_and_auntuncle.newick",
        "justroot.newick",
        "onlychild.newick",
        "triplets.newick",
        "twins.newick",
    ],
)
def test_newick_assets(newick_file: str):
    newick_path = os.path.join(
        os.path.dirname(__file__), "..", "assets", newick_file
    )
    newick = pathlib.Path(newick_path).read_text().strip()
    result = alifestd_from_newick(newick, create_ancestor_list=True)

    assert "id" in result.columns
    assert "ancestor_list" in result.columns
    assert "ancestor_id" in result.columns
    assert "taxon_label" in result.columns

    # root should have ancestor_id == id
    roots = result[result["ancestor_id"] == result["id"]]
    assert len(roots) >= 1

    # all non-root ancestor_ids should reference valid ids
    non_root = result[result["ancestor_id"] != result["id"]]
    assert non_root["ancestor_id"].isin(result["id"]).all()

    # roundtrip: newick -> alife -> newick -> alife preserves topology
    re_newick = alifestd_as_newick_asexual(result, taxon_label="taxon_label")
    re_result = alifestd_from_newick(re_newick)
    assert len(re_result) == len(result)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_roundtrip(phylogeny_df: pd.DataFrame):
    """Test roundtrip: alife -> newick -> alife preserves topology."""
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)

    newick = alifestd_as_newick_asexual(phylogeny_df, taxon_label="id")
    reconstructed = alifestd_from_newick(newick)

    assert len(reconstructed) == len(phylogeny_df)

    # build parent mapping from reconstructed data using taxon_labels as ids
    taxon_labels = dict(zip(reconstructed["id"], reconstructed["taxon_label"]))
    reconstructed_edges = set()
    for _, row in reconstructed.iterrows():
        child_label = row["taxon_label"]
        if row["ancestor_id"] != row["id"]:
            parent_label = taxon_labels[row["ancestor_id"]]
            reconstructed_edges.add((int(child_label), int(parent_label)))
        else:
            reconstructed_edges.add((int(child_label), int(child_label)))

    original_edges = set()
    for _, row in phylogeny_df.iterrows():
        original_edges.add((int(row["id"]), int(row["ancestor_id"])))

    assert reconstructed_edges == original_edges


def test_whitespace_handling():
    result1 = alifestd_from_newick("(A,B)C;")
    result2 = alifestd_from_newick("  (A,B)C;  ")
    assert len(result1) == len(result2)
    assert list(result1["taxon_label"]) == list(result2["taxon_label"])


def test_comments_ignored():
    result = alifestd_from_newick("(A[comment],B)C;")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "A" in labels
    assert "B" in labels
    assert "C" in labels


def test_negative_branch_length():
    result = alifestd_from_newick("(A:-1.5,B:2.0);")
    a = result[result["taxon_label"] == "A"]
    assert a["branch_length"].iloc[0] == pytest.approx(-1.5)


def test_scientific_notation():
    result = alifestd_from_newick("(A:1.5e-3,B:2E4);")
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    assert a["branch_length"].iloc[0] == pytest.approx(1.5e-3)
    assert b["branch_length"].iloc[0] == pytest.approx(2e4)


@pytest.mark.parametrize(
    "phylogeny_csv",
    [
        "example-standard-toy-asexual-phylogeny.csv",
        "example-standard-toy-asexual-phylogeny-noncompact1.csv",
        "example-standard-toy-asexual-phylogeny-noncompact2.csv",
        "example-standard-toy-asexual-phylogeny-uniq.csv",
        "nk_ecoeaselection.csv",
        "nk_lexicaseselection.csv",
        "nk_tournamentselection.csv",
        "prunetestphylo.csv",
        "collapse_unifurcations_testphylo.csv",
    ],
)
@pytest.mark.parametrize("taxon_label", [None, "id"])
@pytest.mark.parametrize("with_branch_length", [True, False])
def test_alifestd_asset_roundtrip(
    phylogeny_csv, taxon_label, with_branch_length
):
    """Test roundtrip: alifestd -> newick -> alifestd using asset files."""
    phylogeny_df = pd.read_csv(f"{assets_path}/{phylogeny_csv}")
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)

    if with_branch_length:
        if (
            "origin_time_delta" not in phylogeny_df.columns
            and "origin_time" not in phylogeny_df.columns
        ):
            phylogeny_df["origin_time_delta"] = np.arange(
                len(phylogeny_df), dtype=float
            )
    else:
        phylogeny_df = phylogeny_df.drop(
            columns=["origin_time", "origin_time_delta"],
            errors="ignore",
        )

    newick = alifestd_as_newick_asexual(
        phylogeny_df,
        taxon_label=taxon_label,
    )
    reconstructed = alifestd_from_newick(newick)

    assert len(reconstructed) == len(phylogeny_df)

    # verify single root
    roots = reconstructed[reconstructed["ancestor_id"] == reconstructed["id"]]
    assert len(roots) == 1

    if taxon_label == "id":
        # verify exact topology via edge matching
        taxon_labels = dict(
            zip(reconstructed["id"], reconstructed["taxon_label"])
        )
        reconstructed_edges = set()
        for _, row in reconstructed.iterrows():
            child_label = row["taxon_label"]
            if row["ancestor_id"] != row["id"]:
                parent_label = taxon_labels[row["ancestor_id"]]
                reconstructed_edges.add((int(child_label), int(parent_label)))
            else:
                reconstructed_edges.add((int(child_label), int(child_label)))

        original_edges = set()
        for _, row in phylogeny_df.iterrows():
            original_edges.add((int(row["id"]), int(row["ancestor_id"])))

        assert reconstructed_edges == original_edges

    if with_branch_length:
        # verify branch lengths are present (not all NaN)
        assert not reconstructed["branch_length"].isna().all()
        assert (
            reconstructed["origin_time_delta"].dropna()
            == reconstructed["branch_length"].dropna()
        ).all()


def test_quoted_label_with_comma():
    result = alifestd_from_newick("('a,b','c,d');")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "a,b" in labels
    assert "c,d" in labels


def test_quoted_label_with_colon():
    result = alifestd_from_newick("('a:1','b:2');")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "a:1" in labels
    assert "b:2" in labels


def test_quoted_label_with_paren():
    result = alifestd_from_newick("('a(b)','c(d)');")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "a(b)" in labels
    assert "c(d)" in labels


def test_quoted_label_with_semicolon():
    result = alifestd_from_newick("('a;b','c;d');")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "a;b" in labels
    assert "c;d" in labels


def test_quoted_label_with_bracket():
    result = alifestd_from_newick("('a[b]','c[d]');")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "a[b]" in labels
    assert "c[d]" in labels


def test_quoted_label_with_branch_length():
    result = alifestd_from_newick("('node A':1.5,'node B':2.5);")
    assert len(result) == 3
    a = result[result["taxon_label"] == "node A"]
    b = result[result["taxon_label"] == "node B"]
    assert a["branch_length"].iloc[0] == pytest.approx(1.5)
    assert b["branch_length"].iloc[0] == pytest.approx(2.5)


def test_quoted_label_with_space_and_branch_length():
    result = alifestd_from_newick(
        "('leaf one':0.1,'leaf two':0.2)'the root':0.0;"
    )
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "leaf one" in labels
    assert "leaf two" in labels
    assert "the root" in labels
    root = result[result["taxon_label"] == "the root"]
    assert root["branch_length"].iloc[0] == pytest.approx(0.0)


def test_nested_comments():
    result = alifestd_from_newick("(A[outer [inner] rest],B)C;")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "A" in labels
    assert "B" in labels
    assert "C" in labels


def test_comment_after_branch_length():
    result = alifestd_from_newick("(A:1.5[comment],B:2.0[another])C;")
    assert len(result) == 3
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    assert a["branch_length"].iloc[0] == pytest.approx(1.5)
    assert b["branch_length"].iloc[0] == pytest.approx(2.0)


def test_comment_before_label():
    result = alifestd_from_newick("([pre]A,[pre]B)C;")
    assert len(result) == 3
    labels = set(result["taxon_label"])
    assert "A" in labels
    assert "B" in labels


def test_branch_length_zero():
    result = alifestd_from_newick("(A:0,B:0.0):0;")
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    assert a["branch_length"].iloc[0] == pytest.approx(0.0)
    assert b["branch_length"].iloc[0] == pytest.approx(0.0)


def test_branch_length_positive_sign():
    result = alifestd_from_newick("(A:+1.5,B:+2.0);")
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    assert a["branch_length"].iloc[0] == pytest.approx(1.5)
    assert b["branch_length"].iloc[0] == pytest.approx(2.0)


def test_root_with_branch_length():
    result = alifestd_from_newick("(A:1,B:2)root:5;")
    root = result[result["taxon_label"] == "root"]
    assert root["branch_length"].iloc[0] == pytest.approx(5.0)


def test_all_internal_unlabeled():
    result = alifestd_from_newick("((A,B),(C,D));")
    assert len(result) == 7
    labeled = result[result["taxon_label"] != ""]
    assert set(labeled["taxon_label"]) == {"A", "B", "C", "D"}
    unlabeled = result[result["taxon_label"] == ""]
    assert len(unlabeled) == 3  # root + two internal


def test_deeply_nested():
    result = alifestd_from_newick("((((((leaf)a)b)c)d)e)root;")
    assert len(result) == 7
    leaf = result[result["taxon_label"] == "leaf"]
    assert len(leaf) == 1
    # trace ancestry: leaf -> a -> b -> c -> d -> e -> root
    cur_id = leaf["id"].iloc[0]
    depth = 0
    while True:
        row = result[result["id"] == cur_id].iloc[0]
        if row["ancestor_id"] == row["id"]:
            break
        cur_id = row["ancestor_id"]
        depth += 1
    assert depth == 6


def test_label_with_underscores_and_digits():
    result = alifestd_from_newick("(node_1:1,node_2_a:2)root_0;")
    labels = set(result["taxon_label"])
    assert "node_1" in labels
    assert "node_2_a" in labels
    assert "root_0" in labels


def test_branch_length_dtype_int():
    result = alifestd_from_newick(
        "(ant:17,(bat:31,cow:22):7,dog:22);",
        branch_length_dtype=int,
    )
    assert result["branch_length"].dtype == pd.Int64Dtype()
    ant = result[result["taxon_label"] == "ant"]
    assert ant["branch_length"].iloc[0] == 17
    # missing branch length should be pd.NA (null)
    root = result[result["ancestor_id"] == result["id"]]
    assert pd.isna(root["branch_length"].iloc[0])


def test_branch_length_dtype_int_no_lengths():
    result = alifestd_from_newick("(A,B)C;", branch_length_dtype=int)
    assert result["branch_length"].dtype == pd.Int64Dtype()
    assert result["branch_length"].isna().all()


def test_star_tree():
    """Multifurcation with many children at root."""
    result = alifestd_from_newick("(A,B,C,D,E,F,G,H);")
    assert len(result) == 9
    root = result[result["ancestor_id"] == result["id"]]
    assert len(root) == 1
    children = result[result["ancestor_id"] != result["id"]]
    assert len(children) == 8
    # all children share the same parent
    assert children["ancestor_id"].nunique() == 1


def test_single_leaf_in_parens():
    result = alifestd_from_newick("(A);")
    assert len(result) == 2
    a = result[result["taxon_label"] == "A"]
    assert len(a) == 1
    root = result[result["ancestor_id"] == result["id"]]
    assert a["ancestor_id"].iloc[0] == root["id"].iloc[0]


def test_empty_label_with_branch_length():
    result = alifestd_from_newick("(:1.0,:2.0):0.5;")
    assert len(result) == 3
    assert (result["taxon_label"] == "").all()
    root = result[result["ancestor_id"] == result["id"]]
    assert root["branch_length"].iloc[0] == pytest.approx(0.5)


def test_spaces_after_comma():
    """Spaces after commas should be stripped, not included in labels."""
    result = alifestd_from_newick("(A, B, C);")
    labels = set(result["taxon_label"])
    assert "A" in labels
    assert "B" in labels
    assert "C" in labels


def test_scientific_notation_positive_exponent():
    result = alifestd_from_newick("(A:1e3,B:2.5E+2);")
    a = result[result["taxon_label"] == "A"]
    b = result[result["taxon_label"] == "B"]
    assert a["branch_length"].iloc[0] == pytest.approx(1e3)
    assert b["branch_length"].iloc[0] == pytest.approx(2.5e2)



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_coarsen_taxa_asexual_make_agg,
    alifestd_make_empty,
)


@pytest.fixture
def sample_phylogeny_df() -> pd.DataFrame:
    # columns to be excluded: id, ancestor_id, ancestor_list
    # override cols: branch_length, destruction_time, edge_length, origin_time
    # other cols: foo, bar
    cols = [
        "id",
        "ancestor_id",
        "ancestor_list",
        "branch_length",
        "destruction_time",
        "edge_length",
        "origin_time",
        "foo",
        "bar",
    ]
    # create an empty DataFrame with just the columns
    return pd.DataFrame({col: [0] for col in cols})


def test_empty_df():
    original_df = alifestd_make_empty(ancestor_id=True)
    result = alifestd_coarsen_taxa_asexual_make_agg(original_df)
    assert isinstance(result, dict)
    assert result == {}


def test_overrides_and_defaults(sample_phylogeny_df):
    agg = alifestd_coarsen_taxa_asexual_make_agg(sample_phylogeny_df)
    # the override columns should map to their override values
    assert agg["destruction_time"] == "last"
    assert agg["origin_time"] == "first"
    # other columns not in overrides should use the default
    assert agg["foo"] == "first"
    assert agg["bar"] == "first"
    # ensure excluded cols are not present
    for excluded in (
        "id",
        "ancestor_id",
        "ancestor_list",
        "branch_length",
        "edge_length",
    ):
        assert excluded not in agg


def test_custom_default(sample_phylogeny_df):
    # if we pass a different default_agg, non-overrides pick that up
    agg = alifestd_coarsen_taxa_asexual_make_agg(
        sample_phylogeny_df,
        default_agg="last",
    )
    # overrides stay the same
    assert "branch_length" not in agg
    assert agg["destruction_time"] == "last"
    assert "edge_length" not in agg
    assert agg["origin_time"] == "first"
    # foo/bar now pick up the new default
    assert agg["foo"] == "last"
    assert agg["bar"] == "last"



import typing

from Bio import Phylo as BioPhylo
from astropy.utils.decorators import deprecated_renamed_argument
import opytional as opyt

from ..._auxiliary_lib import (
    HereditaryStratigraphicInstrument,
    cast_int_lossless,
)
from ._descend_template_phylogeny import descend_template_phylogeny


@deprecated_renamed_argument("seed_column", "seed_instrument", since="1.20.0")
def descend_template_phylogeny_biopython(
    tree: BioPhylo.BaseTree.Tree,
    seed_instrument: HereditaryStratigraphicInstrument,
    extant_nodes: typing.Optional[
        typing.Iterable[BioPhylo.BaseTree.Clade]
    ] = None,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicInstrument]:
    """Generate a population of hereditary stratigraphic instruments that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    tree : biopython Tree
        Phylogeny record as a biopython Tree.
    seed_instrument : HereditaryStratigraphicInstrument
        Hereditary stratigraphic instrument to seed at root node of phylogeny.

        Returned hereditary stratigraphic instrument population will be
        generated as if repeatedly calling `CloneDescendant()` on
        `seed_instrument`. As such, specifies configuration (i.e., differentia
        bit width and stratum retention policy) for returned instruments. May
        already have strata deposited, which will be incorporated into generated
        extant population.
    extant_nodes : optional list of biopython Clade
        Which organisms should hereditary stratigraphic instruments be created
        for?

        Designates content and order of returned list of hereditary
        stratigraphic instrument.

        If None, hereditary stratigraphic instruments will be created for all
        phylogenetic leaves (organisms without offspring) in order of
        appearance in `tree.leaf_node_iter()`.
    progress_wrap : Callable, optional
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicInstrument
        Population of hereditary stratigraphic instruments for extant lineage
        members (i.e., phylogeny leaf nodes).

        Instruments ordered in order of appearance of corresponding extant
        organism id.
    """

    # adapted from https://biopython.org/wiki/Phylo_cookbook
    def all_parents(tree: BioPhylo.BaseTree.Tree) -> typing.Dict:
        parents = {}
        for clade in tree.find_clades(order="level"):
            for child in clade:
                parents[child] = clade
        return parents

    parent_lookup = all_parents(tree)

    return descend_template_phylogeny(
        ascending_lineage_iterables=(
            [extant_node, *tree.trace(extant_node, tree.root)]
            for extant_node in opyt.or_value(
                extant_nodes, tree.get_terminals()
            )
        ),
        descending_tree_iterable=tree.find_clades(),
        get_parent=lambda node: parent_lookup.get(node, None),
        get_stem_length=lambda node: cast_int_lossless(
            opyt.or_value(node.branch_length, 1),
            action="warn",
            context="branch length",
        ),
        seed_instrument=seed_instrument,
        demark=id,
        progress_wrap=progress_wrap,
    )



from ...._auxiliary_lib import get_package_name, load_cppimportable_module

_impl_mod = load_cppimportable_module(
    "_build_tree_searchtable_cpp_impl", get_package_name(__name__)
)
placeholder_value = _impl_mod.placeholder_value
Records = _impl_mod.Records
collapse_unifurcations = _impl_mod.collapse_unifurcations
copy_records_to_dict = _impl_mod.copy_records_to_dict
extract_records_to_dict = _impl_mod.extract_records_to_dict
extend_tree_searchtable_cpp_from_exploded = (
    _impl_mod.extend_tree_searchtable_cpp_from_exploded
)
build_tree_searchtable_cpp_from_exploded = (
    _impl_mod.build_tree_searchtable_cpp_from_exploded
)
build_tree_searchtable_cpp_from_nested = (
    _impl_mod.build_tree_searchtable_cpp_from_nested
)
check_trie_invariant_contiguous_ids = (
    _impl_mod.check_trie_invariant_contiguous_ids
)
check_trie_invariant_topologically_sorted = (
    _impl_mod.check_trie_invariant_topologically_sorted
)
check_trie_invariant_chronologically_sorted = (
    _impl_mod.check_trie_invariant_chronologically_sorted
)
check_trie_invariant_single_root = _impl_mod.check_trie_invariant_single_root
check_trie_invariant_search_children_valid = (
    _impl_mod.check_trie_invariant_search_children_valid
)
check_trie_invariant_search_children_sorted = (
    _impl_mod.check_trie_invariant_search_children_sorted
)
check_trie_invariant_no_indistinguishable_nodes = (
    _impl_mod.check_trie_invariant_no_indistinguishable_nodes
)
check_trie_invariant_data_nodes_are_leaves = (
    _impl_mod.check_trie_invariant_data_nodes_are_leaves
)
check_trie_invariant_search_lineage_compatible = (
    _impl_mod.check_trie_invariant_search_lineage_compatible
)
check_trie_invariant_ancestor_bounds = (
    _impl_mod.check_trie_invariant_ancestor_bounds
)
check_trie_invariant_root_at_zero = _impl_mod.check_trie_invariant_root_at_zero
check_trie_invariant_ranks_nonnegative = (
    _impl_mod.check_trie_invariant_ranks_nonnegative
)
describe_records = _impl_mod._describe_records
diagnose_trie_invariant_contiguous_ids = (
    _impl_mod.diagnose_trie_invariant_contiguous_ids
)
diagnose_trie_invariant_topologically_sorted = (
    _impl_mod.diagnose_trie_invariant_topologically_sorted
)
diagnose_trie_invariant_chronologically_sorted = (
    _impl_mod.diagnose_trie_invariant_chronologically_sorted
)
diagnose_trie_invariant_single_root = (
    _impl_mod.diagnose_trie_invariant_single_root
)
diagnose_trie_invariant_search_children_valid = (
    _impl_mod.diagnose_trie_invariant_search_children_valid
)
diagnose_trie_invariant_search_children_sorted = (
    _impl_mod.diagnose_trie_invariant_search_children_sorted
)
diagnose_trie_invariant_no_indistinguishable_nodes = (
    _impl_mod.diagnose_trie_invariant_no_indistinguishable_nodes
)
diagnose_trie_invariant_data_nodes_are_leaves = (
    _impl_mod.diagnose_trie_invariant_data_nodes_are_leaves
)
diagnose_trie_invariant_search_lineage_compatible = (
    _impl_mod.diagnose_trie_invariant_search_lineage_compatible
)
diagnose_trie_invariant_ancestor_bounds = (
    _impl_mod.diagnose_trie_invariant_ancestor_bounds
)
diagnose_trie_invariant_root_at_zero = (
    _impl_mod.diagnose_trie_invariant_root_at_zero
)
diagnose_trie_invariant_ranks_nonnegative = (
    _impl_mod.diagnose_trie_invariant_ranks_nonnegative
)



import argparse
import logging
import os

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._jit import jit
from ._log_context_duration import log_context_duration


@jit(nopython=True)
def _alifestd_mark_right_child_asexual_fast_path(
    ancestor_ids: np.ndarray,
) -> np.ndarray:
    """Implementation detail for `alifestd_mark_right_child_asexual`."""

    right_children = np.arange(len(ancestor_ids), dtype=ancestor_ids.dtype)
    for idx, ancestor_id in enumerate(ancestor_ids):
        if ancestor_id == idx:
            continue  # handle genesis cases

        cur_right_child = right_children[ancestor_id]
        if cur_right_child == ancestor_id or idx > cur_right_child:
            right_children[ancestor_id] = idx

    return right_children


def _alifestd_mark_right_child_asexual_slow_path(
    phylogeny_df: pd.DataFrame,
) -> pd.DataFrame:
    """Implementation detail for `alifestd_mark_right_child_asexual`."""

    phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["right_child_id"] = phylogeny_df["id"]

    for idx in phylogeny_df.index:
        ancestor_id = phylogeny_df.at[idx, "ancestor_id"]
        if ancestor_id == idx:
            continue  # handle genesis cases

        cur_right_child = phylogeny_df.at[ancestor_id, "right_child_id"]
        if cur_right_child == ancestor_id or idx > cur_right_child:
            phylogeny_df.at[ancestor_id, "right_child_id"] = idx

    return phylogeny_df


def alifestd_mark_right_child_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `right_child`, containing for each node its largest-id child.

    Leaf nodes will be marked with their own id.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df[
            "right_child_id"
        ] = _alifestd_mark_right_child_asexual_fast_path(
            phylogeny_df["ancestor_id"].to_numpy()
        )
        return phylogeny_df
    else:
        return _alifestd_mark_right_child_asexual_slow_path(
            phylogeny_df,
        )


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Add column `right_child`, containing for each node its largest-id child.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_mark_right_child_asexual",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_mark_right_child_asexual",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_mark_right_child_asexual,
            ),
        )



import os

import polars as pl
import pytest

from hstrat.dataframe import surface_validate_trie

assets_path = os.path.join(os.path.dirname(__file__), "assets")


# ---------------------------------------------------------------------------
# Smoke / passthrough tests using trie.csv (from packed.csv via pipeline)
# ---------------------------------------------------------------------------


def test_smoke():
    """Validation passes and returns int for well-formed trie."""
    df = pl.read_csv(f"{assets_path}/trie.csv")
    result = surface_validate_trie(df)
    assert isinstance(result, int)
    assert result == 0


def test_smoke_long():
    """Validation passes for late-divergence trie (lrc=47, mrca=55)."""
    df = pl.read_csv(f"{assets_path}/trie_long.csv")
    result = surface_validate_trie(df)
    assert isinstance(result, int)
    assert result == 0


# ---------------------------------------------------------------------------
# Structural validation failures
# ---------------------------------------------------------------------------


def test_missing_deserialization_columns():
    """Raises if dstream metadata columns are absent."""
    df = pl.read_csv(f"{assets_path}/trie.csv")
    stripped = df.drop("data_hex")
    with pytest.raises(
        ValueError, match="missing downstream metadata columns"
    ):
        surface_validate_trie(stripped)


def test_missing_id_column():
    """Raises if the 'id' column is absent."""
    df = pl.read_csv(f"{assets_path}/trie.csv")
    stripped = df.drop("id")
    with pytest.raises(ValueError, match="missing required column 'id'"):
        surface_validate_trie(stripped)


def test_missing_ancestor_id_column():
    """Raises if the 'ancestor_id' column is absent."""
    df = pl.read_csv(f"{assets_path}/trie.csv")
    stripped = df.drop("ancestor_id")
    with pytest.raises(
        ValueError, match="missing required column 'ancestor_id'"
    ):
        surface_validate_trie(stripped)


def test_non_contiguous_ids():
    """Raises if ids are not contiguous (do not match row indices)."""
    df = pl.read_csv(f"{assets_path}/trie.csv")
    bad = df.with_columns(
        pl.col("id") + 10,
        pl.col("ancestor_id") + 10,
    )
    with pytest.raises(ValueError, match="ids are not contiguous"):
        surface_validate_trie(bad)


def test_no_meta_raises():
    """Trie without dstream metadata columns fails validation."""
    df = pl.read_csv(f"{assets_path}/trie_no_meta.csv")
    with pytest.raises(
        ValueError, match="missing downstream metadata columns"
    ):
        surface_validate_trie(df)


# ---------------------------------------------------------------------------
# MRCA violation detection using trie_long_invalid.csv
# (lrc=47, corrupted mrca_rank=10 → 47 > 10 → violation)
# ---------------------------------------------------------------------------


def test_violation_detected():
    """Violation is detected when MRCA rank is below last retained commonality."""
    df = pl.read_csv(f"{assets_path}/trie_long_invalid.csv")
    num_violations = surface_validate_trie(
        df, max_num_checks=1_000, max_violations=10
    )
    assert num_violations > 0


def test_violation_returns_count():
    """Returns violation count when max_violations is not exceeded."""
    df = pl.read_csv(f"{assets_path}/trie_long_invalid.csv")
    count = surface_validate_trie(df, max_violations=10)
    assert count >= 1


def test_violation_early_return():
    """Returns early once violations exceed max_violations."""
    df = pl.read_csv(f"{assets_path}/trie_long_invalid.csv")
    max_v = 0
    count = surface_validate_trie(df, max_violations=max_v)
    assert count > max_v


# ---------------------------------------------------------------------------
# Parameter tests
# ---------------------------------------------------------------------------


def test_max_num_checks_zero():
    """With max_num_checks=0, no leaf-pair checks are run; returns 0."""
    df = pl.read_csv(f"{assets_path}/trie.csv")
    result = surface_validate_trie(df, max_num_checks=0)
    assert result == 0


def test_max_num_checks_zero_invalid():
    """With max_num_checks=0, even invalid tries return 0 (no checks done)."""
    df = pl.read_csv(f"{assets_path}/trie_long_invalid.csv")
    result = surface_validate_trie(df, max_num_checks=0)
    assert result == 0


def test_seed_reproducibility():
    """Same seed produces same violation count."""
    df = pl.read_csv(f"{assets_path}/trie_long_invalid.csv")
    count_a = surface_validate_trie(df, seed=42, max_violations=10)
    count_b = surface_validate_trie(df, seed=42, max_violations=10)
    assert count_a == count_b


def test_seed_different():
    """Different seeds may yield different counts (probabilistic)."""
    df = pl.read_csv(f"{assets_path}/trie_long_invalid.csv")
    # Both should still detect violations regardless of seed
    count_0 = surface_validate_trie(df, seed=0, max_violations=10)
    count_1 = surface_validate_trie(df, seed=1, max_violations=10)
    assert count_0 >= 1
    assert count_1 >= 1



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_mrca_id_asexual,
    alifestd_to_working_format,
)
from hstrat._auxiliary_lib._alifestd_find_pair_mrca_id_asexual import (
    alifestd_find_pair_mrca_id_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
def test_fuzz_matches_existing(phylogeny_df: pd.DataFrame):
    """Verify pair mrca matches the multi-id mrca function on real data."""
    ids = phylogeny_df["id"].tolist()
    for i in ids[:5]:
        for j in ids[:5]:
            expected = alifestd_find_mrca_id_asexual(
                phylogeny_df, [i, j], mutate=False
            )
            actual = alifestd_find_pair_mrca_id_asexual(
                phylogeny_df, i, j, mutate=False
            )
            assert actual == expected


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    # Tree:  0 -> 1 -> 2, 0 -> 3
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 0],
        }
    )
    original_df = phylogeny_df.copy()

    assert (
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 1, mutate=mutate)
        == 0
    )
    assert (
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 1, 2, mutate=mutate)
        == 1
    )
    assert (
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 2, 3, mutate=mutate)
        == 0
    )
    assert (
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 2, 2, mutate=mutate)
        == 2
    )
    assert (
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 0, mutate=mutate)
        == 0
    )

    if not mutate:
        assert original_df.equals(phylogeny_df)


def test_simple_with_ancestor_list():
    phylogeny_df = alifestd_to_working_format(
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "ancestor_list": ["[None]", "[0]", "[1]", "[0]"],
            }
        )
    )
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 2, 3) == 0
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 1, 2) == 1


def test_single_node():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_id": [0],
        }
    )
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 0) == 0


def test_multiple_roots_disjoint():
    """Two disjoint roots should return None."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 1, 2],
        }
    )
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 1) is None
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 1, 2) is None
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 0) == 0


def test_multiple_roots_partial():
    """Forest: tree1 = {0, 1}, tree2 = {2, 3}."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 2, 2],
        }
    )
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 1) == 0
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 2, 3) == 2
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 2) is None
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 1, 3) is None


def test_chain():
    """Straight chain: 0 -> 1 -> 2 -> 3."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
        }
    )
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 3) == 0
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 1, 3) == 1
    assert alifestd_find_pair_mrca_id_asexual(phylogeny_df, 2, 3) == 2


def test_bypass_kwargs():
    """Test that is_topologically_sorted and has_contiguous_ids kwargs work."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 0],
        }
    )
    result = alifestd_find_pair_mrca_id_asexual(
        phylogeny_df,
        2,
        3,
        is_topologically_sorted=True,
        has_contiguous_ids=True,
    )
    assert result == 0


def test_not_topologically_sorted_raises():
    """Should raise NotImplementedError for unsorted input."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 2, 1, 0],
            "ancestor_id": [2, 0, 0, 0],
        }
    )
    with pytest.raises(NotImplementedError):
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 3)


def test_non_contiguous_ids_raises():
    """Should raise NotImplementedError for non-contiguous ids."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 2, 4],
            "ancestor_id": [0, 0, 2],
        }
    )
    with pytest.raises(NotImplementedError):
        alifestd_find_pair_mrca_id_asexual(phylogeny_df, 0, 4)



import argparse
import logging
import os
import sys

import pandas as pd
from tqdm import tqdm

from ._alifestd_collapse_unifurcations import alifestd_collapse_unifurcations
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_leaves import alifestd_mark_leaves
from ._alifestd_to_working_format import alifestd_to_working_format
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


def alifestd_test_leaves_isomorphic_asexual(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    taxon_label: str,
    mutate: bool = False,
    progress_wrap: callable = lambda x: x,
) -> bool:
    """Test if phylogenetic relationships between leaf nodes are topologically
    isomorphic between two phylogenies."""

    if not mutate:
        df1, df2 = df1.copy(), df2.copy()

    if taxon_label == "id":
        df1["taxon_label"], df2["taxon_label"] = df1["id"], df2["id"]
        taxon_label = "taxon_label"

    df1 = alifestd_to_working_format(df1, mutate=True)
    df2 = alifestd_to_working_format(df2, mutate=True)

    df1 = alifestd_collapse_unifurcations(df1, mutate=True)
    df2 = alifestd_collapse_unifurcations(df2, mutate=True)

    df1 = alifestd_to_working_format(df1, mutate=True)
    df2 = alifestd_to_working_format(df2, mutate=True)

    df1 = alifestd_mark_leaves(df1, mutate=True)
    df2 = alifestd_mark_leaves(df2, mutate=True)

    df1, df2 = df1.reset_index(drop=True), df2.reset_index(drop=True)
    assert alifestd_has_contiguous_ids(df1)
    assert alifestd_has_contiguous_ids(df2)

    if len(df1) != len(df2):
        return False

    leaves1, leaves2 = df1[df1["is_leaf"]], df2[df2["is_leaf"]]

    if set(leaves1[taxon_label]) != set(leaves2[taxon_label]):
        return False

    df2_id_lookup = dict(zip(leaves2[taxon_label], leaves2["id"]))
    id_map = dict(zip(leaves1["id"], leaves1[taxon_label].map(df2_id_lookup)))

    for id1 in progress_wrap(df1["id"][::-1]):
        assert id1 in id_map
        ancestor_id1 = df1["ancestor_id"].iat[id1]

        id2 = id_map[id1]
        ancestor_id2 = df2["ancestor_id"].iat[id2]

        if ancestor_id1 in id_map:
            if id_map[ancestor_id1] != ancestor_id2:
                return False
        else:
            id_map[ancestor_id1] = ancestor_id2

    return True


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()})

Test if phylogenetic relationships between leaf nodes are topologically
isomorphic between two phylogenies.

Return code 0 indicates isomorphism; 1 indicates non-isomorphism.

Note that this CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "first_phylogeny",
        type=str,
        help="Path to first alife-standard phylogeny file",
    )
    parser.add_argument(
        "second_phylogeny",
        type=str,
        help="Path to second alife-standard phylogeny file",
    )
    parser.add_argument(
        "-l",
        "--taxon-label",
        type=str,
        help="Name of column to use as taxon label.",
        required=True,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args = parser.parse_args()
    input_ext = [1]

    logging.info(f"reading first phylogeny from {args.first_phylogeny}...")
    first_df = {
        ".csv": pd.read_csv,
        ".fea": pd.read_feather,
        ".feather": pd.read_feather,
        ".pqt": pd.read_parquet,
        ".parquet": pd.read_parquet,
    }[os.path.splitext(args.first_phylogeny)[1]](args.first_phylogeny)

    logging.info(f"reading second phylogeny from {args.second_phylogeny}...")
    second_df = {
        ".csv": pd.read_csv,
        ".fea": pd.read_feather,
        ".feather": pd.read_feather,
        ".pqt": pd.read_parquet,
        ".parquet": pd.read_parquet,
    }[os.path.splitext(args.second_phylogeny)[1]](args.second_phylogeny)

    with log_context_duration(
        "hstrat._auxiliary_lib.alifestd_test_leaves_isomorphic_asexual",
        logging.info,
    ):
        result = alifestd_test_leaves_isomorphic_asexual(
            first_df,
            second_df,
            taxon_label=args.taxon_label,
            progress_wrap=tqdm,
        )

    exit_code = [1, 0][result]

    logging.info(f"exiting with return code {exit_code} for result {result}")
    sys.exit(exit_code)



import typing

from interval_search import binary_search

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from ...genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratigraphicSurface,
)
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_rank_of_last_retained_commonality_between_generic import (
    calc_rank_of_last_retained_commonality_between_generic,
)


def calc_rank_of_last_retained_commonality_between_bsearch(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find rank of strata commonality before first strata disparity.

    Implementation detail. Provides optimized implementation for
    special case where both first and second use the perfect resolution
    stratum retention policy.
    """

    if isinstance(first, HereditaryStratigraphicSpecimen) or isinstance(
        second, HereditaryStratigraphicSpecimen
    ):
        raise ValueError(
            f"HereditaryStratigraphicSpecimen not supported for {calc_rank_of_last_retained_commonality_between_bsearch.__name__}"
        )

    # both must have (effectively) used the perfect resolution policy
    assert not first.HasDiscardedStrata() and not second.HasDiscardedStrata()

    lower_bound = 0
    upper_bound = min(
        [
            first.GetNumStrataDeposited() - 1,
            second.GetNumStrataDeposited() - 1,
        ]
    )
    assert lower_bound <= upper_bound

    def differentia_at(
        which: typing.Union[
            HereditaryStratigraphicColumn, HereditaryStratigraphicSurface
        ],
        idx: int,
    ) -> int:
        return which.GetStratumAtColumnIndex(idx).GetDifferentia()

    def predicate(idx: int) -> bool:
        return differentia_at(first, idx) != differentia_at(second, idx)

    first_disparite_idx = binary_search(
        predicate,
        lower_bound,
        upper_bound,
    )

    collision_implausibility_threshold = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
        first,
        second,
        significance_level=1.0 - confidence_level,
    )
    assert collision_implausibility_threshold > 0
    if first_disparite_idx is None:
        # no disparate strata found
        # fall back to calc_rank_of_last_retained_commonality_between_generic
        # to handle proper bookkeeping in this case while skipping most of
        # the search
        return calc_rank_of_last_retained_commonality_between_generic(
            first,
            second,
            first_start_idx=upper_bound,
            second_start_idx=upper_bound,
            confidence_level=confidence_level,
        )
    elif first_disparite_idx >= collision_implausibility_threshold:
        # disparate strata found, following some common strata
        # ...discount collision_implausibility_threshold - 1 common strata
        # as potential spurious differentia collisions
        # ... must also subtract 1 (canceling out -1 above) to account for
        # moving from disparite stratum to preceding common stratum
        last_common_idx = (
            first_disparite_idx - collision_implausibility_threshold
        )
        last_common_rank = first.GetRankAtColumnIndex(
            last_common_idx,
        )
        assert last_common_idx == last_common_rank
        return last_common_rank
    else:
        # no common strata between first and second
        # or not enough common strata to discount the possibility all
        # are spurious collisions with respect to the given confidence
        # level; conservatively conclude there is no common ancestor
        return None



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_find_leaf_ids,
    alifestd_find_root_ids,
    alifestd_has_multiple_roots,
    alifestd_make_balanced_bifurcating,
    alifestd_make_comb,
    alifestd_make_empty,
    alifestd_mark_sackin_index_asexual,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    original = phylogeny_df.copy()

    result = alifestd_mark_sackin_index_asexual(phylogeny_df)

    assert alifestd_validate(result)
    assert original.equals(phylogeny_df)

    # Sackin index should be non-negative
    assert all(result["sackin_index"] >= 0)

    # Leaf nodes should have Sackin index 0
    leaf_ids = [*alifestd_find_leaf_ids(phylogeny_df)]
    for leaf_id in leaf_ids:
        sackin = result[result["id"] == leaf_id]["sackin_index"].squeeze()
        assert sackin == 0

    # Root should have the highest or equal Sackin index
    if not alifestd_has_multiple_roots(phylogeny_df):
        (root_id,) = alifestd_find_root_ids(phylogeny_df)
        root_sackin = result[result["id"] == root_id]["sackin_index"].squeeze()
        assert root_sackin >= 0
        # Root sackin should be >= all other nodes' sackin
        assert root_sackin == result["sackin_index"].max()


def test_empty():
    res = alifestd_mark_sackin_index_asexual(alifestd_make_empty())
    assert "sackin_index" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_chain(mutate: bool):
    """Test a simple chain/caterpillar tree: 0 -> 1 -> 2.

    Sackin index accumulates through all nodes.
    Node 1: sackin = 0 + 1 = 1 (from leaf 2)
    Node 0: sackin = 1 + 1 = 2 (from node 1)
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaf has Sackin = 0
    assert result_df.loc[2, "sackin_index"] == 0

    # Node 1: 0 + 1 = 1
    assert result_df.loc[1, "sackin_index"] == 1

    # Node 0: 1 + 1 = 2
    assert result_df.loc[0, "sackin_index"] == 2

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_bifurcating_balanced(mutate: bool):
    r"""Test a balanced bifurcating tree.

          0
         / \
        1   2

    Node 0 has 2 children, each with 1 leaf.
    Sackin at root = (0 + 1) + (0 + 1) = 2
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[2, "sackin_index"] == 0

    # Root: sum of (child_sackin + child_leaves) = (0+1) + (0+1) = 2
    assert result_df.loc[0, "sackin_index"] == 2

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple_bifurcating_imbalanced(mutate: bool):
    r"""Test an imbalanced bifurcating tree.

          0
         / \
        1   2
           / \
          3   4

    Leaves: 1, 3, 4 (depths from root: 1, 2, 2)
    Node 2: sackin = (0+1) + (0+1) = 2
    Node 0: sackin = (0+1) + (2+2) = 5

    Total Sackin at root = 1 + 2 + 2 = 5
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[None]", "[0]", "[0]", "[2]", "[2]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[3, "sackin_index"] == 0
    assert result_df.loc[4, "sackin_index"] == 0

    # Node 2: (0+1) + (0+1) = 2
    assert result_df.loc[2, "sackin_index"] == 2

    # Node 0: (0+1) + (2+2) = 5
    assert result_df.loc[0, "sackin_index"] == 5

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_caterpillar_tree(mutate: bool):
    r"""Test a caterpillar/comb tree (maximally imbalanced).

          0
         / \
        1   2
           / \
          3   4
             / \
            5   6

    Leaves: 1 (depth 1), 3 (depth 2), 5 (depth 3), 6 (depth 3)
    Node 4: (0+1) + (0+1) = 2
    Node 2: (0+1) + (2+2) = 5
    Node 0: (0+1) + (5+3) = 9

    Total Sackin = 1 + 2 + 3 + 3 = 9
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[2]",
                "[2]",
                "[4]",
                "[4]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[3, "sackin_index"] == 0
    assert result_df.loc[5, "sackin_index"] == 0
    assert result_df.loc[6, "sackin_index"] == 0

    # Node 4: (0+1) + (0+1) = 2
    assert result_df.loc[4, "sackin_index"] == 2

    # Node 2: (0+1) + (2+2) = 5
    assert result_df.loc[2, "sackin_index"] == 5

    # Node 0: (0+1) + (5+3) = 9
    assert result_df.loc[0, "sackin_index"] == 9

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_polytomy_balanced(mutate: bool):
    r"""Test a tree with balanced polytomy (more than 2 children).

          0
        / | \
       1  2  3

    Sackin index handles polytomies.
    Node 0: (0+1) + (0+1) + (0+1) = 3
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[0]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[2, "sackin_index"] == 0
    assert result_df.loc[3, "sackin_index"] == 0

    # Root with 3 children: (0+1) + (0+1) + (0+1) = 3
    assert result_df.loc[0, "sackin_index"] == 3

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_polytomy_imbalanced(mutate: bool):
    r"""Test a tree with imbalanced polytomy.

            0
          / | \
         1  2  3
              / \
             4   5

    Node 3: (0+1) + (0+1) = 2
    Node 0: (0+1) + (0+1) + (2+2) = 6

    Leaves: 1 (depth 1), 2 (depth 1), 4 (depth 2), 5 (depth 2)
    Total Sackin = 1 + 1 + 2 + 2 = 6
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": ["[None]", "[0]", "[0]", "[0]", "[3]", "[3]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[2, "sackin_index"] == 0
    assert result_df.loc[4, "sackin_index"] == 0
    assert result_df.loc[5, "sackin_index"] == 0

    # Node 3: (0+1) + (0+1) = 2
    assert result_df.loc[3, "sackin_index"] == 2

    # Node 0: (0+1) + (0+1) + (2+2) = 6
    assert result_df.loc[0, "sackin_index"] == 6

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_large_polytomy(mutate: bool):
    r"""Test a tree with large polytomy (4 children).

            0
         / | | \
        1  2 3  4

    All leaves at depth 1.
    Node 0: (0+1) + (0+1) + (0+1) + (0+1) = 4
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[None]", "[0]", "[0]", "[0]", "[0]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # All leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[2, "sackin_index"] == 0
    assert result_df.loc[3, "sackin_index"] == 0
    assert result_df.loc[4, "sackin_index"] == 0

    # Root with 4 children: 4 * 1 = 4
    assert result_df.loc[0, "sackin_index"] == 4

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_large_polytomy_imbalanced(mutate: bool):
    r"""Test a tree with large imbalanced polytomy.

              0
         / | | \
        1  2 3  4
               / \
              5   6

    Node 4: (0+1) + (0+1) = 2
    Node 0: (0+1) + (0+1) + (0+1) + (2+2) = 7

    Leaves: 1,2,3 (depth 1), 5,6 (depth 2)
    Total Sackin = 1 + 1 + 1 + 2 + 2 = 7
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[0]",
                "[0]",
                "[4]",
                "[4]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[2, "sackin_index"] == 0
    assert result_df.loc[3, "sackin_index"] == 0
    assert result_df.loc[5, "sackin_index"] == 0
    assert result_df.loc[6, "sackin_index"] == 0

    # Node 4: (0+1) + (0+1) = 2
    assert result_df.loc[4, "sackin_index"] == 2

    # Node 0: (0+1) + (0+1) + (0+1) + (2+2) = 7
    assert result_df.loc[0, "sackin_index"] == 7

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_non_contiguous_ids(mutate: bool):
    """Test with non-contiguous IDs to exercise slow path."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [10, 20, 30, 40, 50],
            "ancestor_list": ["[None]", "[10]", "[10]", "[30]", "[30]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Same structure as test_simple_bifurcating_imbalanced
    # Leaves have Sackin = 0
    assert result_df.loc[20, "sackin_index"] == 0
    assert result_df.loc[40, "sackin_index"] == 0
    assert result_df.loc[50, "sackin_index"] == 0

    # Node 30: (0+1) + (0+1) = 2
    assert result_df.loc[30, "sackin_index"] == 2

    # Node 10: (0+1) + (2+2) = 5
    assert result_df.loc[10, "sackin_index"] == 5

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_non_contiguous_ids_polytomy(mutate: bool):
    """Test polytomy with non-contiguous IDs to exercise slow path."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [10, 20, 30, 40, 50, 60],
            "ancestor_list": [
                "[None]",
                "[10]",
                "[10]",
                "[10]",
                "[40]",
                "[40]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Node 40: (0+1) + (0+1) = 2
    assert result_df.loc[40, "sackin_index"] == 2

    # Node 10: (0+1) + (0+1) + (2+2) = 6
    assert result_df.loc[10, "sackin_index"] == 6

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_multiple_roots(mutate: bool):
    """Test with multiple roots (forest)."""
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Two separate trees, each with one child (unifurcating)
    # Generalized handles unifurcations
    assert result_df.loc[2, "sackin_index"] == 0  # leaf
    assert result_df.loc[3, "sackin_index"] == 0  # leaf
    assert result_df.loc[0, "sackin_index"] == 1  # 0 + 1 from child 2
    assert result_df.loc[1, "sackin_index"] == 1  # 0 + 1 from child 3

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_larger_imbalanced_tree(mutate: bool):
    r"""Test a larger imbalanced tree.

              0
             / \
            1   2
               / \
              3   4
                 / \
                5   6
                   / \
                  7   8

    Leaves: 1 (depth 1), 3 (depth 2), 5 (depth 3), 7 (depth 4), 8 (depth 4)
    Node 6: (0+1) + (0+1) = 2
    Node 4: (0+1) + (2+2) = 5
    Node 2: (0+1) + (5+3) = 9
    Node 0: (0+1) + (9+4) = 14

    Total Sackin = 1 + 2 + 3 + 4 + 4 = 14
    """
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[2]",
                "[2]",
                "[4]",
                "[4]",
                "[6]",
                "[6]",
            ],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_sackin_index_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]

    # Leaves have Sackin = 0
    assert result_df.loc[1, "sackin_index"] == 0
    assert result_df.loc[3, "sackin_index"] == 0
    assert result_df.loc[5, "sackin_index"] == 0
    assert result_df.loc[7, "sackin_index"] == 0
    assert result_df.loc[8, "sackin_index"] == 0

    # Node 6: (0+1) + (0+1) = 2
    assert result_df.loc[6, "sackin_index"] == 2

    # Node 4: (0+1) + (2+2) = 5
    assert result_df.loc[4, "sackin_index"] == 5

    # Node 2: (0+1) + (5+3) = 9
    assert result_df.loc[2, "sackin_index"] == 9

    # Node 0: (0+1) + (9+4) = 14
    assert result_df.loc[0, "sackin_index"] == 14

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_df, expected_sackin",
    [
        # R treebalance::sackinI and treestats::sackin reference values
        # 2-leaf balanced: (A,B) -> Sackin = 2
        (alifestd_make_balanced_bifurcating(2), 2),
        # 4-leaf balanced: ((A,B),(C,D)) -> Sackin = 8
        (alifestd_make_balanced_bifurcating(3), 8),
        # 8-leaf balanced: (((A,B),(C,D)),((E,F),(G,H))) -> Sackin = 24
        (alifestd_make_balanced_bifurcating(4), 24),
        # 3-leaf caterpillar: (A,(B,C)) -> Sackin = 5
        (alifestd_make_comb(3), 5),
        # 4-leaf caterpillar: (A,(B,(C,D))) -> Sackin = 9
        (alifestd_make_comb(4), 9),
        # 5-leaf caterpillar: (A,(B,(C,(D,E)))) -> Sackin = 14
        (alifestd_make_comb(5), 14),
        # 7-leaf caterpillar -> Sackin = 27
        (alifestd_make_comb(7), 27),
    ],
)
def test_against_r_treebalance_sackin(
    phylogeny_df: pd.DataFrame, expected_sackin: int
):
    """Test Sackin index against values computed with R treebalance::sackinI
    and treestats::sackin packages (both agreed on all values)."""
    result_df = alifestd_mark_sackin_index_asexual(phylogeny_df)
    result_df.index = result_df["id"]
    root_id = result_df[result_df["id"] == result_df["ancestor_id"]][
        "id"
    ].iloc[0]
    assert result_df.loc[root_id, "sackin_index"] == expected_sackin


@pytest.mark.parametrize(
    "phylogeny_df, expected_sackin",
    [
        # R treebalance::sackinI reference values for nontrivial trees
        # (neither purely balanced nor purely comb/caterpillar).
        # ((A,B),(C,(D,E))) -> Sackin = 12
        (
            pd.DataFrame(
                {
                    "id": range(9),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[6]",
                        "[6]",
                    ],
                }
            ),
            12,
        ),
        # ((A,(B,C)),(D,E)) -> Sackin = 12
        (
            pd.DataFrame(
                {
                    "id": range(9),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[4]",
                        "[4]",
                    ],
                }
            ),
            12,
        ),
        # (((A,B),C),((D,E),F)) -> Sackin = 16
        (
            pd.DataFrame(
                {
                    "id": range(11),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[3]",
                        "[3]",
                        "[5]",
                        "[5]",
                    ],
                }
            ),
            16,
        ),
        # (((A,B),(C,D)),(E,F)) -> Sackin = 16
        (
            pd.DataFrame(
                {
                    "id": range(11),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[3]",
                        "[3]",
                        "[4]",
                        "[4]",
                    ],
                }
            ),
            16,
        ),
        # ((A,(B,(C,D))),(E,(F,G))) -> Sackin = 21
        (
            pd.DataFrame(
                {
                    "id": range(13),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[4]",
                        "[4]",
                        "[6]",
                        "[6]",
                        "[8]",
                        "[8]",
                    ],
                }
            ),
            21,
        ),
        # (((A,B),(C,(D,E))),((F,G),(H,I))) -> Sackin = 29
        (
            pd.DataFrame(
                {
                    "id": range(17),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[3]",
                        "[3]",
                        "[4]",
                        "[4]",
                        "[5]",
                        "[5]",
                        "[6]",
                        "[6]",
                        "[10]",
                        "[10]",
                    ],
                }
            ),
            29,
        ),
        # ((A,B),((C,D),((E,F),(G,H)))) -> Sackin = 26
        (
            pd.DataFrame(
                {
                    "id": range(15),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[5]",
                        "[5]",
                        "[6]",
                        "[6]",
                        "[9]",
                        "[9]",
                        "[10]",
                        "[10]",
                    ],
                }
            ),
            26,
        ),
        # Polytomy: (A,B,(C,(D,E))) -> Sackin = 10
        (
            pd.DataFrame(
                {
                    "id": range(8),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[0]",
                        "[3]",
                        "[3]",
                        "[5]",
                        "[5]",
                    ],
                }
            ),
            10,
        ),
        # Polytomy: ((A,B,C),(D,(E,F))) -> Sackin = 14
        (
            pd.DataFrame(
                {
                    "id": range(10),
                    "ancestor_list": [
                        "[None]",
                        "[0]",
                        "[0]",
                        "[1]",
                        "[1]",
                        "[1]",
                        "[2]",
                        "[2]",
                        "[7]",
                        "[7]",
                    ],
                }
            ),
            14,
        ),
    ],
)
def test_against_r_nontrivial_trees_sackin(
    phylogeny_df: pd.DataFrame, expected_sackin: int
):
    """Test Sackin index against R treebalance::sackinI values for nontrivial
    trees (neither purely balanced nor purely comb/caterpillar)."""
    result_df = alifestd_mark_sackin_index_asexual(phylogeny_df)
    result_df.index = result_df["id"]
    root_id = result_df[result_df["id"] == result_df["ancestor_id"]][
        "id"
    ].iloc[0]
    assert result_df.loc[root_id, "sackin_index"] == expected_sackin



import argparse
import functools
import gc
import logging
import os

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import pandas as pd

from ._add_bool_arg import add_bool_arg
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_topological_sensitivity_warned import (
    alifestd_topological_sensitivity_warned,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@alifestd_topological_sensitivity_warned(
    insert=False,
    delete=True,
    update=True,
)
def alifestd_delete_trunk_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Delete entries masked by `is_trunk` column.

    Masked entries must be contiguous, meaning that no non-trunk entry can
    be an ancestor of a trunk entry. Children of deleted entries will become
    roots.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_collapse_trunk_asexual
    """
    if "is_trunk" not in phylogeny_df:
        raise ValueError(
            "`is_trunk` column not provided, trunk is unspecified"
        )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    has_contiguous_ids = bool(alifestd_has_contiguous_ids(phylogeny_df))
    logging.info(f"- alifestd_delete_trunk_asexual: {has_contiguous_ids=}")
    if has_contiguous_ids:
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    logging.info(
        "- alifestd_delete_trunk_asexual: marking ancestor_is_trunk...",
    )
    if has_contiguous_ids:
        phylogeny_df["ancestor_is_trunk"] = phylogeny_df[
            "is_trunk"
        ].to_numpy()[phylogeny_df["ancestor_id"]]
    else:
        phylogeny_df["ancestor_is_trunk"] = phylogeny_df.loc[
            phylogeny_df["ancestor_id"], "is_trunk"
        ].to_numpy()

    logging.info("- alifestd_delete_trunk_asexual: testing special cases...")
    if np.any(phylogeny_df["is_trunk"] & ~phylogeny_df["ancestor_is_trunk"]):
        raise ValueError("specified trunk is non-contiguous")

    if phylogeny_df["is_trunk"].sum() == 0:
        return phylogeny_df

    if "ancestor_id" in phylogeny_df:
        logging.info(
            "- alifestd_delete_trunk_asexual: updating ancestor_id...",
        )
        phylogeny_df.loc[
            phylogeny_df["ancestor_is_trunk"], "ancestor_id"
        ] = phylogeny_df.loc[
            phylogeny_df["ancestor_is_trunk"], "id"
        ].to_numpy()

    if "ancestor_list" in phylogeny_df:
        logging.info(
            "- alifestd_delete_trunk_asexual: updating ancestor_list...",
        )
        phylogeny_df.loc[
            phylogeny_df["ancestor_is_trunk"], "ancestor_list"
        ] = "[none]"

    gc.collect()

    logging.info("- alifestd_delete_trunk_asexual: filtering should_keep...")
    should_keep = ~phylogeny_df["is_trunk"]
    res = phylogeny_df.loc[should_keep].reset_index(drop=True)

    assert res["is_trunk"].sum() == 0
    return res


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Delete entries masked by `is_trunk` column.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_delete_trunk_asexual",
        dfcli_version=get_hstrat_version(),
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_delete_trunk_asexual", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                functools.partial(
                    alifestd_delete_trunk_asexual,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            ),
        )



import typing

from astropy.utils.decorators import deprecated_renamed_argument
from iterpop import iterpop as ip
import networkx as nx
import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicInstrument, unpairwise
from ._descend_template_phylogeny import descend_template_phylogeny


@deprecated_renamed_argument("seed_column", "seed_instrument", since="1.20.0")
def descend_template_phylogeny_networkx(
    tree: nx.DiGraph,
    seed_instrument: HereditaryStratigraphicInstrument,
    extant_nodes: typing.Optional[typing.Iterable[typing.Any]] = None,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicInstrument]:
    """Generate a population of hereditary stratigraphic instruments that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    tree : nx.DiGraph
        Phylogeny record as a networkx DiGraph.

        Assumes directed edges point from ancestor to descendant. Does not
        support non-unit edge length.
    seed_instrument : HereditaryStratigraphicInstrument
        Hereditary stratigraphic instrument to seed at root node of phylogeny.

        Returned hereditary stratigraphic instrument population will be
        generated as if repeatedly calling `CloneDescendant()` on
        `seed_instrument`. As such, specifies configuration (i.e., differentia
        bit width and stratum retention policy) for returned instruments. May
        already have strata deposited, which will be incorporated into generated
        extant population.
    extant_nodes : optional list of networkx nodes
        Which organisms should hereditary stratigraphic instruments be created
        for?

        Designates content and order of returned list of hereditary
        stratigraphic instrument.

        If None, hereditary stratigraphic instruments will be created for all
        phylogenetic leaves (organisms without offspring) in order of
        appearance in `tree.nodes`.
    progress_wrap : Callable, optional
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicInstrument
        Population of hereditary stratigraphic instruments for extant lineage
        members (i.e., phylogeny leaf nodes).

        Instruments ordered in order of appearance of corresponding extant
        organism id.
    """

    # handle empty case
    if len(tree) == 0:
        return []

    reversed_tree = tree.reverse(copy=False)
    root_node = ip.popsingleton(
        node for node in tree.nodes if tree.in_degree(node) == 0
    )
    leaf_nodes = [node for node in tree.nodes if tree.out_degree(node) == 0]

    return descend_template_phylogeny(
        ascending_lineage_iterables=(
            unpairwise(nx.dfs_edges(reversed_tree, extant_node))
            if len(tree) > 1
            else (extant_node,)
            for extant_node in opyt.or_value(extant_nodes, leaf_nodes)
        ),
        descending_tree_iterable=(
            unpairwise(nx.dfs_edges(tree, root_node))
            if len(tree) > 1
            else (root_node,)
        ),
        get_parent=lambda node: ip.poursingleton(reversed_tree[node]),
        get_stem_length=lambda node: 1,
        seed_instrument=seed_instrument,
        demark=lambda x: x,
        progress_wrap=progress_wrap,
    )



#!/usr/bin/env python3

from hstrat import hstrat

if __name__ == "__main__":

    for policy in (
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(3),
    ):
        print()
        print(policy)

        col = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=policy,
        )
        for __ in range(20):
            col.DepositStratum()

        print(hstrat.col_to_ascii(col))

    for policy in (
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(3),
    ):
        print()
        print(policy)

        cols = [
            hstrat.HereditaryStratigraphicColumn(
                stratum_retention_policy=policy,
            )
            for __ in range(3)
        ]
        for col in cols:
            for __ in range(20):
                col.DepositStratum()

        assemblage = hstrat.pop_to_assemblage(cols)
        print(hstrat.assemblage_to_ascii(assemblage))



import typing

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_empty,
    alifestd_mark_num_preceding_leaves_asexual,
    alifestd_to_working_format,
)


def test_empty():
    res = alifestd_mark_num_preceding_leaves_asexual(alifestd_make_empty())
    assert len(res) == 0


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(apply: typing.Callable, mutate: bool):
    # Chain tree: 0 -> 1 -> 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_mark_num_preceding_leaves_asexual(
            phylogeny_df,
            mutate=mutate,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_num_preceding_leaves_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[3, "num_preceding_leaves"] == 0
    assert result_df.loc[0, "num_preceding_leaves"] == 0
    assert result_df.loc[1, "num_preceding_leaves"] == 0
    assert result_df.loc[2, "num_preceding_leaves"] == 2
    assert result_df.loc[4, "num_preceding_leaves"] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_num_preceding_leaves_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "num_preceding_leaves"] == 0
    assert result_df.loc[1, "num_preceding_leaves"] == 0
    assert result_df.loc[2, "num_preceding_leaves"] == 2
    assert result_df.loc[3, "num_preceding_leaves"] == 0
    assert result_df.loc[4, "num_preceding_leaves"] == 1
    assert result_df.loc[5, "num_preceding_leaves"] == 2
    assert result_df.loc[6, "num_preceding_leaves"] == 3

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple5(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df.copy()
    result_df = alifestd_mark_num_preceding_leaves_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "num_preceding_leaves"] == 0
    assert result_df.loc[1, "num_preceding_leaves"] == 0
    assert result_df.loc[2, "num_preceding_leaves"] == 3
    assert result_df.loc[3, "num_preceding_leaves"] == 0
    assert result_df.loc[4, "num_preceding_leaves"] == 1
    assert result_df.loc[5, "num_preceding_leaves"] == 1
    assert result_df.loc[6, "num_preceding_leaves"] == 2


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple6(apply: typing.Callable, mutate: bool):
    # Chain tree: 0
    phylogeny_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[None]"],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_num_preceding_leaves_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "num_preceding_leaves"] == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple7(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0             7
    #       /   \          / \
    #      1     2        8   9
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
                "[None]",
                "[7]",
                "[7]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_num_preceding_leaves_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "num_preceding_leaves"] == 0
    assert result_df.loc[1, "num_preceding_leaves"] == 0
    assert result_df.loc[2, "num_preceding_leaves"] == 3
    assert result_df.loc[3, "num_preceding_leaves"] == 0
    assert result_df.loc[4, "num_preceding_leaves"] == 1
    assert result_df.loc[5, "num_preceding_leaves"] == 1
    assert result_df.loc[6, "num_preceding_leaves"] == 2
    assert result_df.loc[7, "num_preceding_leaves"] == 0
    assert result_df.loc[8, "num_preceding_leaves"] == 0
    assert result_df.loc[9, "num_preceding_leaves"] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple8(mutate: bool):
    # Chain tree: 0
    phylogeny_df = pd.DataFrame(
        {
            "id": [1],
            "ancestor_list": ["[None]"],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_num_preceding_leaves_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "num_preceding_leaves"] == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)
